""" Script that launches a server.

Normally, this is meant to be invoked from within a Launcher object (see launcher.py).
However, you can also call this directly, with command-line arguments:
:arg --logport: port number of log server
:arg --serverport: (optional) port number to use for server. Notes:
    (1) if not provided, the user will be prompted to enter a port number in the commandline
    (2) this script does not explicitly provide ConnectionRefusedError handling. Two (non-exclusive) options:
        (i) Implement in server.launch() method
        (ii) Implement in launcher.py module (DONE)
:arg --server: (optional) the name of the server module. Notes:
    (1) should be the name of the module from which the desired server can be instantiated
    (2) module must explicitly be imported in this script
            (we could consider adding auto-import functionality for selected/all modules at a later time)
    (3) module must have a launch(**kwargs) method for instantiating server (and prerequisites) with parameters
        :param logger: instance of LogClient class to use for logging
        :param port: (int) port number
"""

import importlib

from pylabnet.utils.helper_methods import parse_args, show_console, hide_console
from pylabnet.utils.logging.logger import LogClient


def main():

    # parse command line arguments
    args = parse_args()
    try:
        log_port = int(args['logport'])
    except IndexError:
        raise IndexError('Please provide command line arguments in the form\n"'
                         'python launch_gui.py --logport 1234 --serverport 5678 --server servername')
    if 'serverport' in args:
        server_port = int(args['serverport'])
    else:
        show_console()
        server_port = int(input('Please enter a server port value: '))
        hide_console()
    if 'server' in args:
        server = args['server']
    else:
        show_console()
        server = input('Please enter a server module name: ')
        hide_console()
    if 'logip' in args:
        host = args['logip']
    else:
        host = 'localhost'

    # Instantiate logger
    server_logger = LogClient(
        host=host,
        port=log_port,
        module_tag=server+'_server',
        server_port=server_port
    )

    # Retrieve debug flag.
    debug = int(args['debug'])

    # Halt execution and wait for debugger connection if debug flag is up.
    if debug:
        import ptvsd
        import os
        # 5678 is the default attach port in the VS Code debug configurations
        server_logger.info(f"Waiting for debugger to attach to PID {os.getpid()} (pylabnet_server)")
        ptvsd.enable_attach(address=('localhost', 5678))
        ptvsd.wait_for_attach()
        breakpoint()

    # Instantiate module
    try:
        mod_inst = importlib.import_module(f'servers.{server}')
    except ModuleNotFoundError:
        server_logger.error(f'No module found in pylabnet.launchers.servers named {server}.py')
        raise

    mod_inst.launch(logger=server_logger, port=server_port)


if __name__ == '__main__':
    main()



import numpy as np
import copy


class PulseBlock:
    """ Class for construction of pulse sequences.

    PulseBlock is essentially a container which has several "shelves" (channels),
    and each pulse is represented by a "box" (Pulse object) sitting on the shelf.
    Each "box" must have a name tag: "channel name - start time - duration".
    PulseBlock makes no additional assumptions about the contents of the boxes.

    Since the "boxes" do not necessarily cover the entire duration, there are
    some gaps. To specify what is happening during the gaps, one uses DfltPulse
    objects - "default pulses" (a single DfltPulse per channel).

    PulseBlock handles everything related to keeping the "boxes" time-ordered
    when new elements are added: it has methods for inserting additional "boxes"
    into arbitrary (empty) places on the "shelf" and for merging several smaller
    PulseBlocks into a large one.


    --------------- Structure ---------------

    The main attribute of PulseBlock object is p_dict - the pulse dictionary.
    This is where the entire pulse sequence is stored. In the dictionary:
        Key - the channel name;
        Value - time-ordered list of Pulse objects for this channel.

    After any operation, time axis origin is shifted for the entire PulseBlock
    such that the beginning of earliest Pulse is set to 0.

    PulseBlock assumes that each Pulse object has the following attributes:
        ch - (str) channel name;
        t0 - (numeric) pulse start time;
        dur - (numeric) duration of the pulse,

    dflt_dict attribute contains the default pulse objects:
        Key - (str) channel name
        Value - DfltPulse object

    In addition, PulseBlock has the following attributes:
        name (str) - name of the sequence
        dur (numeric) - total duration of the sequence
                        (t0+dur of the latest Pulse object)
    """

    def __init__(self, p_obj_list=None, dflt_dict=None, name=''):
        """ Construct new PulseBlock.

        Each argument is optional. One can pass non-default values to fill-in
        some of the attributes during construction, but any pulses can be added
        into any location later. Name can also be changed later.

        :param p_obj_list: (opt) list of Pulse objects.
            This argument is used to insert some pulses immediately at instantiation.
            Example:
                p_obj_list = [
                    pulse.PTrue(ch='ch1', t0=0, dur=1e-6),
                    pulse.PSin(ch='ch2', t0=1e-6, dur=1e-6, amp=1e-3, freq=2.5e9, ph=0)
                ]

        :param dflt_dict: (opt) dictionary of default-pulse values.
            Keys - channel names, values - DfltPulse objects
            Can be filled-in at any point later - optional at instantiation.
            Example:
                dflt_dict = {
                    'aom': pulse.DFalse()
                    'analog_out': pulse.DConst(val=0.0)
                }

        :param name: (opt, str) pulse object name
        """

        self.name = name
        self.dur = 0
        self.p_dict = dict()
        self.dflt_dict = dict()

        if dflt_dict is not None:
            self.dflt_dict = copy.deepcopy(dflt_dict)

        if p_obj_list is None:
            p_obj_list = list()

        # Iterate through all given pulses and _insert them into the block
        # (_insert should be used - it does not reset edges)
        for p_obj in p_obj_list:
            p_obj = copy.deepcopy(p_obj)
            self._insert(p_obj=p_obj)

        # Reset edges: set the left-most edge to zero
        # and set self.dur to the right-most edge
        self.reset_edges()

    def _insert(self, p_obj, cflct_er=True):
        """ Technical method for inserting a new Pulse object into PulseBlock
        Here start and stop edges of PulseBlock are not adjusted.

        Insertion location is specified by Pulse object p_obj attributes:
            ch - which channel to insert into. If the specified channel is not
                 yet present in the PulseBlock, the new channel is registered.

            t0 - time to place the beginning of the p_obj with respect to the
                 beginning of PulseBlock. If t0 is negative, p_obj is inserted
                 before the earliest pulse in PulseBlock. Time origin is NOT
                 adjusted - p_obj will have negative t0.

        :param p_obj: the Pulse object to be inserted

        :param cflct_er: (bool) 'conflict-error' check.
                         If True, before actually inserting p_obj, a check is
                         performed to ensure that p_obj does not overlap with
                         existing pulses on the channel. In the case of overlap,
                         PulseBlock is not altered and ValueError is produced.
        :return: None
        """

        p_obj = copy.deepcopy(p_obj)
        ch = p_obj.ch

        # Sanity check:
        #   new pulse does not conflict with existing pulses
        if cflct_er and ch in self.p_dict.keys():

            p_list = self.p_dict[ch]

            # Find the position, into which
            # the new pulse should be inserted
            t0_list = np.array(
                [p_item.t0 for p_item in p_list]
            )
            idx = np.searchsorted(t0_list, p_obj.t0)

            # If the new pulse will not be the left-most [idx = 0],
            # check for overlap with existing pulse to the left
            if idx > 0:
                if not (p_list[idx - 1].t0 + p_list[idx - 1].dur) <= p_obj.t0:
                    cflct_item = p_list[idx - 1]

                    raise ValueError(
                        'insert(): conflict on ch="{}":  given pulse \n'
                        '   {}, t0={:.2e}, dur={:.2e} \n'
                        'overlaps with existing pulse to the left \n'
                        '   {}, t0={:.2e}, dur={:.2e}'
                        ''.format(
                            ch,
                            str(p_obj),
                            p_obj.t0,
                            p_obj.dur,
                            str(cflct_item),
                            cflct_item.t0,
                            cflct_item.dur
                        )
                    )

            # If the new pulse will not be the right-most [idx = len-1],
            # check for overlap with existing pulse to the right
            if idx < len(p_list):
                if not (p_obj.t0 + p_obj.dur) <= p_list[idx].t0:
                    cflct_item = p_list[idx]

                    raise ValueError(
                        'insert(): conflict on ch="{}": given pulse \n'
                        '   {} t0={:.2e}, dur={:.2e} \n'
                        'overlaps with existing pulse to the right \n'
                        '   {} t0={:.2e}, dur={:.2e}'
                        ''.format(
                            ch,
                            str(p_obj),
                            p_obj.t0,
                            p_obj.dur,
                            str(cflct_item),
                            cflct_item.t0,
                            cflct_item.dur
                        )
                    )

        # Create a new entry for 'ch' if it is not yet registered
        if ch not in self.p_dict.keys():
            self.p_dict[ch] = []

        # Add p_obj as a new entry in 'ch' pulse list
        self.p_dict[ch].append(p_obj)

        # T-order pulses within 'ch'
        self.p_dict[ch].sort(
            key=lambda p_item: p_item.t0
        )

    def insert(self, p_obj, cflct_er=True):
        """ Insert a new Pulse object into PulseBlock

        Insertion location is specified by Pulse object p_obj attributes:
            ch - which channel to insert into. If the specified channel is not
                 yet present in the PulseBlock, the new channel is registered.

            t0 - time to place the beginning of the p_obj with respect to the
                 beginning of PulseBlock. If t0 is negative, p_obj is inserted
                 before the earliest pulse in PulseBlock and time origin is
                 shifted into the beginning of p_obj.

        :param p_obj: the Pulse object to be inserted

        :param cflct_er: (bool) 'conflict-error' check.
                         If True, before actually inserting p_obj, a check is
                         performed to ensure that p_obj does not overlap with
                         existing pulses on the channel. In the case of overlap,
                         PulseBlock is not altered and ValueError is produced.
        :return: None
        """

        self._insert(p_obj=p_obj, cflct_er=cflct_er)
        self.reset_edges()

    def join(self, p_obj, name='', cflct_er=True):
        """ Same as insert(), but instead of changing the existing PulseBlock,
        a new one is created and the original PulseBlock is not altered.

        :param name: name of the new PulseBlock object
        :return: (PulseBlock) the new PulseBlock - self with p_obj inserted.
        """

        new_pb = copy.deepcopy(self)
        new_pb.name = name
        new_pb.insert(
            p_obj=p_obj,
            cflct_er=cflct_er
        )

        return new_pb

    def append(self, p_obj, join=False, name=None, cflct_er=True):
        """ Same as insert()/join(), but t0 of p_obj now refers to the position
        with respect to the end of PulseBlock.

        :param join: (bool) - if False, the existing PulseBlock is modified
                              [insert() method is called, name param is not used]
                              if True, new PulseBlock with name 'name'
                              is created [join() method is called]
        """

        p_obj = copy.deepcopy(p_obj)
        p_obj.t0 += self.dur

        if join:

            if name is None:
                name = self.name

            return self.join(
                p_obj=p_obj,
                name=name,
                cflct_er=cflct_er
            )

        else:
            self.insert(
                p_obj=p_obj,
                cflct_er=cflct_er
            )

    def insert_pb(self, pb_obj, t0=0, cflct_er=True):
        """ Insert PulseBlock pb_obj into self.

        :param pb_obj: the PulseBlock object to be inserted

        :param t0: position where the beginning of pb_obj should be placed with
                   respect to the beginning of self.
                   If t0 is negative, pb_obj is inserted before the beginning of
                   self and time origin is shifted into the beginning of pb_obj.

        :param cflct_er: (bool) If True, conflict check is performed across all
                         channels (for both Pulse and DfltPulse values).

        :return: None
        """

        # TODO: re-implement using reset_edges()

        pb_obj = copy.deepcopy(pb_obj)

        #
        # Sanity checks
        #

        # No overlap between blocks
        if cflct_er and (set(self.p_dict.keys()) & set(pb_obj.p_dict.keys())):
            if t0 >= 0:
                if not self.dur <= t0:
                    raise ValueError(
                        'insert_pb(): blocks overlap and cannot be merged: \n'
                        '   t0={:.2e} does not exceed self.dur={:.2e}'
                        ''.format(t0, self.dur)
                    )
            else:
                if not pb_obj.dur <= abs(t0):
                    raise ValueError(
                        'insert_pb(): blocks overlap and cannot be merged: \n'
                        '   offset -t0={:.2e} is smaller than pb_obj.dur={:.2e}'
                        ''.format(-t0, pb_obj.dur)
                    )
        # No conflicts between default pulse objects
        if cflct_er:
            for ch in pb_obj.dflt_dict.keys():
                if ch in self.dflt_dict.keys() and pb_obj.dflt_dict[ch] != self.dflt_dict[ch]:
                    raise ValueError(
                        'insert_pb(): conflict between default pulse objects on channel "{}": \n'
                        '   self.dflt_dict[{}] = {} \n'
                        ' pb_obj.dflt_dict[{}] = {}'
                        ''.format(
                            ch,
                            ch,
                            str(self.dflt_dict[ch]),
                            ch,
                            str(pb_obj.dflt_dict[ch])
                        )
                    )

        #
        # Insert pb
        #

        # Calculate duration
        if t0 >= 0:
            self.dur = max(self.dur, pb_obj.dur + t0)
        else:
            self.dur = max(abs(t0) + self.dur, pb_obj.dur)

        # Shift all elements of the right-most block
        if t0 >= 0:
            for ch in pb_obj.p_dict.keys():
                for p_item in pb_obj.p_dict[ch]:
                    p_item.t0 += t0
        else:
            for ch in self.p_dict.keys():
                for p_item in self.p_dict[ch]:
                    p_item.t0 += abs(t0)

        # Register new channels
        for ch in pb_obj.p_dict.keys():
            if ch not in self.p_dict.keys():
                self.p_dict[ch] = []

        # Add new pulses
        for ch in pb_obj.p_dict.keys():
            self.p_dict[ch].extend(
                pb_obj.p_dict[ch]
            )

            if t0 < 0:
                # Pulses from pb_obj were added from the right
                # Sort pulses in T-order to move them to the left
                self.p_dict[ch].sort(
                    key=lambda pulse_item: pulse_item.t0
                )

        # Add new default pulse objects
        for ch in pb_obj.dflt_dict.keys():
            if ch not in self.dflt_dict.keys():
                self.dflt_dict[ch] = copy.deepcopy(
                    pb_obj.dflt_dict[ch]
                )

    def join_pb(self, pb_obj, t0=0, cflct_er=True, name=''):
        """ Same as insert_pb(), but instead of modifying self,
        a new PulseBlock is created. Self is not altered.

        :return: (PulseBlock) new PulseBlock - self with pb_obj inserted.
        """

        new_pb = copy.deepcopy(self)
        new_pb.name = name
        new_pb.insert_pb(
            pb_obj=pb_obj,
            t0=t0,
            cflct_er=cflct_er
        )
        return new_pb

    def append_pb(self, pb_obj, offset=0, join=False):
        """ Same as insert_pb()/join_pb(), but now the position for the beginning
        of pb_obj to be placed into should be given with respect to the end of self.

        :param offset: the position into which the beginning of pb_obj should be
                       placed with respect to the end of self.

                       NOTE: only NON-negative offset is allowed.
                       To achieve negative offset, use insert_pb()/join_pb()
        """

        if offset < 0:
            raise ValueError(
                'append_pb(): only non-negative offset is allowed. \n'
                'Use insert_pb() or join_pb() for arbitrary t0'
            )

        if join:
            return self.join_pb(
                pb_obj=pb_obj,
                t0=self.dur + offset,
                name=self.name
            )
        else:
            self.insert_pb(
                pb_obj=pb_obj,
                t0=self.dur + offset
            )

    def reset_edges(self):

        p_ch_list = list(self.p_dict.keys())
        if len(p_ch_list) == 0:
            self.dur = 0
            return

        # Left edge ==============

        # Find the left-most edge - it will be set to origin
        left_edge_list = [
            self.p_dict[ch][0].t0 for ch in p_ch_list
        ]
        left_edge = min(left_edge_list)

        # Shift all pulses such that the left-most edge
        # coincides with zero
        for ch in p_ch_list:
            for p_item in self.p_dict[ch]:
                p_item.t0 -= left_edge

        # Right edge ==============

        # Find the right-most edge - the duration of the block
        right_edge_list = [
            self.p_dict[ch][-1].t0 + self.p_dict[ch][-1].dur for ch in p_ch_list
        ]
        self.dur = max(right_edge_list)

    def __str__(self):
        """ Print PulseBlock

        :return: (str) formatted printout of PulseBlock
        """

        pulse_ch_list = sorted(list(self.p_dict.keys()))
        dflt_ch_list = sorted(list(self.dflt_dict.keys()))

        ret_str = 'PulseBlock "{}" \n' \
                  'pulse_chs = {} \n' \
                  'deflt_chs = {} \n' \
                  'dur = {:.2e} \n' \
                  'p_dict: \n' \
                  ''.format(self.name, pulse_ch_list, dflt_ch_list, self.dur)

        for ch in self.p_dict.keys():
            ch_str = '    {}: \n'.format(ch)
            for p_obj in self.p_dict[ch]:
                ch_str += '        {{{:.2e}, {:.2e}, {}}}\n'.format(p_obj.t0, p_obj.dur, str(p_obj))

            ret_str += ch_str
            ret_str += '\n'

        ret_str += 'dflt_dict: \n'
        for ch in self.dflt_dict.keys():
            ret_str += '    {}: {} \n'.format(ch, str(self.dflt_dict[ch]))

        return ret_str

    # Experiment-oriented convenience utils

    def ch_map(self, map_dict):
        """ Rename channels.

        This method is supposed to be used when high-level, human-readable
        channel names need to be mapped onto hardware-specific channel name
        strings.
        For example: 'mw_drive' -> 'AO1', 'AOM_gate' -> 'DO1'

        :param map_dict: (dict) Name mapping dictionary:
                         Key - current name of the channel in PulseObject;
                         Value - the new name.

                         For example:
                            map_dict = {'mw_drive': 'AO1', 'AOM_gate': 'DO1'}
        :return: None.
        """

        #
        # Sanity check:
        #

        # - map_dict.keys() covers p_dict.keys()
        if not set(self.p_dict.keys()) <= set(map_dict.keys()):
            raise ValueError(
                'ch_map(): map_dict does not include all channels of p_dict: \n'
                '   map_dict.keys()={} \n'
                'self.p_dict.keys()={} \n'
                ''.format(
                    sorted(map_dict.keys()),
                    sorted(self.p_dict.keys())
                )
            )

        # - map_dict.keys() covers dflt_dict.keys()
        if not set(self.dflt_dict.keys()) <= set(map_dict.keys()):
            raise ValueError(
                'ch_map(): map_dict does not include all channels of dflt_dict \n'
                '      map_dict.keys() = {} \n'
                'self.dlft_dict.keys() = {}'
                ''.format(
                    sorted(map_dict.keys()),
                    sorted(self.dflt_dict.keys())
                )
            )

        #
        # Channel mapping
        #

        # Store references to the original pulse
        # dictionaries in temporary variables
        tmp_p_dict = self.p_dict
        tmp_dflt_dict = self.dflt_dict

        # Empty p_dict and dflt_dict
        self.p_dict = dict()
        self.dflt_dict = dict()

        # Fill-in new p_dict with original Pulse lists
        for ch in tmp_p_dict.keys():
            self.p_dict[map_dict[ch]] = tmp_p_dict[ch]

        # Fill-in new dflt_dict with original DfltPulse values
        for ch in tmp_dflt_dict.keys():
            self.dflt_dict[map_dict[ch]] = tmp_dflt_dict[ch]

        # Rename 'ch' attribute in every Pulse object
        for new_ch in self.p_dict.keys():
            for p_item in self.p_dict[new_ch]:
                p_item.ch = new_ch

        del tmp_p_dict, tmp_dflt_dict

    def add_offset(self, offset_dict):
        """ Time-offset several channel.

        This method is supposed to be used to offset all pulses on a given
        channel/channels with respect to the rest of PulseBlock.

        This is normally used when one needs to compensate for the physical delay
        between channels.

        :param offset_dict: (dict) offset dictionary:
                            Key - channel name
                            Value - offset (both positive and negative
                            values are acceptable)
        :return: None.
        """

        # Sanity check: all given channels are present in p_dict

        if not set(offset_dict.keys()) <= set(self.p_dict.keys()):
            raise ValueError(
                'offset dictionary contains unknown channels: \n'
                'offset_dict.keys() = {} \n'
                'self.p_dict.keys() = {}'
                ''.format(
                    sorted(offset_dict.keys()),
                    sorted(self.p_dict.keys())
                )
            )

        for ch in offset_dict.keys():
            offset = offset_dict[ch]

            for p_item in self.p_dict[ch]:
                p_item.t0 += offset

        self.reset_edges()



from pyvisa import VisaIOError, ResourceManager

from pylabnet.utils.logging.logger import LogHandler
import numpy as np
import time


class Driver():
    """Driver class for GPIB controlled Agilent EE405 Spectrum analyser"""

    def reset(self):
        """ Create factory reset"""
        self.device.write('*RST')
        self.log.info("Reset to factory settings successfull.")

    def __init__(self, gpib_address, logger):
        """Instantiate driver class

        :gpib_address: GPIB-address of spectrum analyzer, e.g. 'GPIB0::12::INSTR'
            Can be read out by using
                rm = pyvisa.ResourceManager()
                rm.list_resources()
        :logger: And instance of a LogClient
        """

        # Instantiate log
        self.log = LogHandler(logger=logger)

        self.rm = ResourceManager()

        try:
            self.device = self.rm.open_resource(gpib_address)
            device_id = self.device.query('*IDN?')
            self.log.info(f"Successfully connected to {device_id}.")
        except VisaIOError:
            self.log.error(f"Connection to {gpib_address} failed.")

        # reset to factory settings
        self.reset()

    def display_off(self):
        """ Power off display """
        self.device.write(':DISPlay:ENABle OFF')
        self.log.info("Display off.")

    def display_on(self):
        """ Power on display """
        self.device.write(':DISPlay:ENABle ON')
        self.log.info("Display on.")

    def set_attenuation(self, db):
        """ Set input attenuation
        :db: Target attenuation in dB, must be between 0 and 75
        """

        if not 0 <= db <= 75:
            self.log.error(
                f'Invalid attenuation ({db}dB). Attenuation must be between 0dB and 75dB'
            )
        self.device.write(f'POW:ATT {int(db)}dB')
        self.log.info(f'Input attenuation set to {db}dB.')

    def set_center_frequency(self, center_frequency):
        """ Set center frequency of trace.

        :center_frequency: Frequency in Hz (from 0 to 13.3 GHz)
        """
        if not 0 <= center_frequency <= 13.2*1e9:
            self.log.error(
                f'Invalid center frequency ({center_frequency} Hz). Must be within 0 and 13.2 GHz'
            )

        self.device.write(f':SENSe:FREQuency:CENTer {center_frequency}')
        self.log.info(f'Center frequency set to {center_frequency} Hz')

    def set_frequency_span(self, frequency_span):
        """ Set frequency span of trace.

        :frequency_span: Frequency span in Hz (from 0 to 13.3 GHz)
        """
        if not 0 <= frequency_span <= 13.2*1e9:
            self.log.error(
                f'Invalid frequency span ({frequency_span} Hz). Must be within 0 and 13.2 GHz'
            )

        self.device.write(f':SENSe:FREQuency:SPAN {frequency_span}')
        self.log.info(f'Frequency span set {frequency_span} Hz')

    def toggle_cont(self, target_state):
        """Switch between single shot and continuous acquisition mode

        :target_state: Index of targe stat. 1 for continuous mode, 0 for single shot mode.
        """
        self.device.write(f'INIT:CONT {target_state}')

    def get_frequency_array(self):
        """Constructs array of frequencies associated with trace points"""

        # Sweep start frequency.
        start_freq = float(self.device.query(':SENSe:FREQuency:STARt?'))

        # Sweep end frequency.
        end_freq = float(self.device.query(':SENSe:FREQuency:STOP?'))

        # Number of sweep points.
        num_sweep_points = float(self.device.query('SENSE:SWEEP:POINTS?'))

        # Array containing frequencies of each sweep point.
        frequencies = np.linspace(start_freq, end_freq, num_sweep_points)

        return frequencies

    def read_trace(self):
        """ Read and return trace

        Retruns array trace contaning frequencies (in Hz) of data points in
        trace[:,0] and power levels (in dBm) in trace[:,1]
        """

        # Set to single shot mode.
        self.toggle_cont(0)

        # Trigger a sweep and wait for sweep to complete.
        self.device.write('INIT:IMM;*WAI')

        # Specify units in dBm.
        self.device.write('UNIT:POW DBM')

        # Specify data format as ASCII.
        self.device.write('FORM:DATA ASC')

        # Trigger a sweep and wait for sweep to complete.
        self.device.write('INIT:IMM;*WAI')

        # Query trace data
        dbm_measurement = self.device.query_ascii_values('TRAC:DATA? TRACE1')

        # Return to continuos monitoring mode
        self.toggle_cont(1)

        # Read frequency axis
        frequencies = self.get_frequency_array()

        # Combine trace data
        trace = np.stack((frequencies, dbm_measurement), axis=-1)

        return trace

    def acquire_background_spectrum(self, num_points=100):
        """Acquires an average background trace.

        :num_traces: How many traces to sample.
        :nocheck: Boolean indicating if user confirmation that no
            input signal is present will be omitted.
        """

        traces = []
        self.display_off()
        for i in range(num_points):
            traces.append(self.read_trace())
        self.display_on()

        noise_average = np.average(np.asarray(traces)[:, :, 1], 0)
        noise_trace = np.array([traces[0][:, 0], noise_average])

        self.noise_trace = np.transpose(noise_trace)

    def get_background(self):

        try:
            noise_trace = self.noise_trace
            return noise_trace
        except AttributeError:
            error_msg = 'No background acquired. Run acquire_background_spectrum() to acquire background.'
            self.log.warn(error_msg)
            return error_msg


class E4405BMarker():
    """ Class handling assignment and read out of peak markers."""

    def __init__(self, e4405Bclient, name, marker_num):
        """ Initialized the marker

        :e4405Bclient: Instance of E4405BClient.
        :name: A human readable name of the marker.
        :marker_num: Marker number from 1 to 4.
        """
        self.client = e4405Bclient
        self.marker_num = marker_num
        self.name = name

        # Set marker as position marker (floating, unassigned to a peak).
        self.set_as_position()

    def set_as_position(self):
        """Specify marker type as position marker (standard)."""
        self.client.write(f':CALCulate:MARKer{self.marker_num}:MODE POSition;*WAI')

    def set_to_maximum(self):
        """Set marker to maximum peak value of trace."""
        self.client.write(f':CALCulate:MARKer{self.marker_num}:MAXimum;*WAI')

    def _toggle_freq_count(self, state):
        """Change frequency count setting

        :state (int): 1 to toggle frequency counting on, 0 to toggle it off.
        """
        self.client.write(f':CALCulate:MARKer{self.marker_num}:FCOunt:STATe {state};*WAI')

    def _get_freq(self):
        """ Read out frequency

        Note: Work only if _toggle_freq_count was called, outputs 9e15 if count state is off.
        """
        return float(self.client.query(f':CALCulate:MARKer{self.marker_num}:FCOunt:X?;*WAI'))

    def get_power(self):
        """Reads out power of marker position (in dbm)"""
        return float(self.client.query(f':CALCulate:MARKer{self.marker_num}:Y?;*WAI'))

    def look_left(self):
        """Focus marker on next peak left."""
        self.client.write(f':CALCulate:MARKer{self.marker_num}:MAXimum:LEFT;*WAI')
        time.sleep(0.1)

    def look_right(self):
        """Focus marker on next peak right."""
        self.client.write(f':CALCulate:MARKer{self.marker_num}:MAXimum:RIGHT;*WAI')
        time.sleep(0.1)

    def read_freq(self):
        """ Read out frequency of maker.

        This is done by first enabling count state, reading out the frequency
        and disabling the count state. For some reason, very long sleep times
        between commands are necessary to ensure correct readout of the frequency.
        """

        # Turn on frequency capture
        self._toggle_freq_count(1)

        time.sleep(1)

        # Read off frequency
        freq = self._get_freq()

        time.sleep(1)

        # Turn off frequency capture
        self._toggle_freq_count(0)

        return freq



import pickle

from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Service(ServiceBase):

    def exposed_assign_plot(self, plot_widget, plot_label, legend_widget):
        return self._module.assign_plot(
            plot_widget=plot_widget,
            plot_label=plot_label,
            legend_widget=legend_widget
        )

    def exposed_clear_plot(self, plot_widget):
        return self._module.clear_plot(
            plot_widget=plot_widget
        )

    def exposed_assign_curve(self, plot_label, curve_label, error=False):
        return self._module.assign_curve(
            plot_label=plot_label,
            curve_label=curve_label,
            error=error
        )

    def exposed_remove_curve(self, plot_label, curve_label):
        self._module.remove_curve(
            plot_label=plot_label,
            curve_label=curve_label
        )

    def exposed_assign_scalar(self, scalar_widget, scalar_label):
        return self._module.assign_scalar(
            scalar_widget=scalar_widget,
            scalar_label=scalar_label
        )

    def exposed_assign_label(self, label_widget, label_label):
        return self._module.assign_label(
            label_widget=label_widget,
            label_label=label_label
        )

    def exposed_assign_event_button(self, event_widget, event_label):
        return self._module.assign_event_button(
            event_widget=event_widget,
            event_label=event_label,

        )

    def exposed_assign_container(self, container_widget, container_label):
        return self._module.assign_container(self, container_widget, container_label)

    def exposed_set_curve_data(self, data_pickle, plot_label, curve_label, error_pickle=None):
        data = pickle.loads(data_pickle)
        error = pickle.loads(error_pickle)
        return self._module.set_curve_data(
            data=data,
            plot_label=plot_label,
            curve_label=curve_label,
            error=error
        )

    def exposed_set_scalar(self, value_pickle, scalar_label):
        value = pickle.loads(value_pickle)
        return self._module.set_scalar(
            value=value,
            scalar_label=scalar_label
        )

    def exposed_get_scalar(self, scalar_label):
        return pickle.dumps(self._module.get_scalar(scalar_label))

    def exposed_activate_scalar(self, scalar_label):
        return self._module.activate_scalar(scalar_label)

    def exposed_deactivate_scalar(self, scalar_label):
        return self._module.deactivate_scalar(scalar_label)

    def exposed_set_label(self, text, label_label):
        return self._module.set_label(
            text=text,
            label_label=label_label
        )

    def exposed_get_text(self, label_label):
        return pickle.dumps(self._module.get_text(label_label))

    def exposed_was_button_pressed(self, event_label):
        return self._module.was_button_pressed(event_label)

    def exposed_change_button_background_color(self, event_label, color):
        return self._module.change_button_background_color(self, event_label, color)

    def exposed_get_container_info(self, container_label):
        return pickle.dumps(self._module.get_container_info(container_label))


class Client(ClientBase):

    def assign_plot(self, plot_widget, plot_label, legend_widget):
        return self._service.exposed_assign_plot(
            plot_widget=plot_widget,
            plot_label=plot_label,
            legend_widget=legend_widget
        )

    def clear_plot(self, plot_widget):
        return self._service.exposed_clear_plot(
            plot_widget=plot_widget
        )

    def assign_curve(self, plot_label, curve_label, error=False):
        return self._service.exposed_assign_curve(
            plot_label=plot_label,
            curve_label=curve_label,
            error=error
        )

    def remove_curve(self, plot_label, curve_label):
        return self._service.exposed_remove_curve(
            plot_label=plot_label,
            curve_label=curve_label
        )

    def assign_scalar(self, scalar_widget, scalar_label):
        self._service.exposed_assign_scalar(
            scalar_widget=scalar_widget,
            scalar_label=scalar_label
        )

    def assign_label(self, label_widget, label_label):
        return self._service.exposed_assign_label(
            label_widget=label_widget,
            label_label=label_label
        )

    def assign_event_button(self, event_widget, event_label):
        return self._service.exposed_assign_event_button(
            event_widget=event_widget,
            event_label=event_label,
        )

    def assign_container(self, container_widget, container_label):
        return self._service.exposed_assign_container(container_widget, container_label)

    def set_curve_data(self, data, plot_label, curve_label, error=None):
        data_pickle = pickle.dumps(data)
        error_pickle = pickle.dumps(error)
        return self._service.exposed_set_curve_data(
            data_pickle=data_pickle,
            plot_label=plot_label,
            curve_label=curve_label,
            error_pickle=error_pickle
        )

    def set_scalar(self, value, scalar_label):
        value_pickle = pickle.dumps(value)
        return self._service.exposed_set_scalar(
            value_pickle=value_pickle,
            scalar_label=scalar_label
        )

    def get_scalar(self, scalar_label):
        return pickle.loads(self._service.exposed_get_scalar(scalar_label))

    def activate_scalar(self, scalar_label):
        return self._service.exposed_activate_scalar(scalar_label)

    def deactivate_scalar(self, scalar_label):
        return self._service.exposed_deactivate_scalar(scalar_label)

    def set_label(self, text, label_label):
        return self._service.exposed_set_label(
            text=text,
            label_label=label_label
        )

    def get_text(self, label_label):
        return pickle.loads(self._service.exposed_get_text(label_label))

    def was_button_pressed(self, event_label):
        return self._service.exposed_was_button_pressed(event_label)

    def change_button_background_color(self, event_label, color):
        return self._service.exposed_change_button_background_color(self, event_label, color)

    def get_container_info(self, container_label):
        return pickle.loads(self._service.exposed_get_container_info(container_label))




import codecs
import os.path
from setuptools import setup, find_packages

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pylabnet',
    version=get_version('pylabnet/__init__.py'),
    description='Client-server, python-based laboratory software',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/lukingroup/pylabnet',
    author='Lukin SiV Team',
    author_email='b16lukin@gmail.com',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.7',  # This may not be strictly necessary
    entry_points={
        'console_scripts': [
            'pylabnet=pylabnet.launchers.launch_control:main',
            'pylabnet_proxy=pylabnet.launchers.launch_control:main_proxy'
            ]
        },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    install_requires=[
        'decorator>=4.4.0',
        'IPython>=7.12.0',
        'ipywidgets>=7.5.1',
        'matplotlib>=3.1.3',
        'nidaqmx>=0.5.7',
        'numpy>=1.16.5',
        'plotly>=4.7.1'
        'ptvsd>=4.3.2',
        'PyQt5>=5.13.0',
        'pyqtgraph>=0.10.0',
        'pyserial>=3.4',
        'pytz>=2019.3',
        'PyVISA>=1.10.1',
        'rpyc>=4.1.2',
        'zhinst>=20.1.1211'
    ]
)



# -*- coding: utf-8 -*-

"""
This file contains the pylabnet Hardware class
for the Zurich Instruments HDAWG.
"""

import zhinst.utils
import re
import time
import textwrap
import copy

from pylabnet.utils.logging.logger import LogHandler

from pylabnet.utils.decorators.logging_redirector import log_standard_output

# Storing the sampling rates and the corresponding target integers for the setInt command
SAMPLING_RATE_DICT = {
    '2.4 GHz':       0,
    '1.2 GHz':       1,
    '600 MHz':       2,
    '300 MHz':       3,
    '150 MHz':       4,
    '75 MHz':        5,
    '37.5 MHz':      6,
    '18.75 MHz':     7,
    '9.37 MHz':      8,
    '4.68 MHz':      9,
    '2.34 MHz':      10,
    '1.17 MHz':      11,
    '585.93 kHz':    12,
    '292.96 kHz':    13,
}


class Driver():

    def reset_DIO_outputs(self):
        """Sets all DIO outputs to low"""
        self.seti('dios/0/output', 0)
        self.log.info("Set all DIO outputs to low.")

    def disable_everything(self):
        """ Create a base configuration.
        Disable all available outputs, awgs, demods, scopes, etc.
        """
        zhinst.utils.disable_everything(self.daq, self.device_id)
        self.log.info("Disabled everything.")

    @log_standard_output
    def log_stdout(self, function):
        """ Execute function and log print output to self.log

        This statement is needed for an inline call where any zhinst command is
        executed and the standard output should be logged
        :function: The function to be executed.
         """
        return function()

    def _convert_to_list(self, input_argument):
        """Checks if input is list and if not, converts it to list."""
        if type(input_argument) is not list:
            input_argument = [input_argument]
        return input_argument

    def __init__(self, device_id, logger, api_level=6):
        """ Instantiate AWG

        :logger: instance of LogClient class
        :device_id: Device id of connceted ZI HDAWG, for example 'dev8060'
        :api_level: API level of zhins API
        """

        # Instantiate log
        self.log = LogHandler(logger=logger)

        # Part of this code has been modified from
        # ZI's Zurich Instruments LabOne Python API Example

        # Call a zhinst utility function that returns:
        # - an API session `daq` in order to communicate
        # with devices via the data server.
        # - the device ID string that specifies the device
        # branch in the server's node hierarchy.
        # - the device's discovery properties.

        err_msg = "This example can only be run on an HDAWG."

        # Connect to device and log print output, not the lambda expression.
        (daq, device, props) = self.log_stdout(
            lambda: zhinst.utils.create_api_session(
                device_id,
                api_level,
                required_devtype='HDAWG',
                required_err_msg=err_msg
            )
        )

        self.log_stdout(lambda: zhinst.utils.api_server_version_check(daq))

        self.daq = daq
        self.device_id = device

        # Create a base configuration
        self.disable_everything()
        self.reset_DIO_outputs()

        # read out number of channels from property dictionary
        self.num_outputs = int(
            re.compile('HDAWG(4|8{1})').match(props['devicetype']).group(1)
        )

    @log_standard_output
    def seti(self, node, new_int):
        """
        Warapper for daq.setInt commands. For instance, instead of
        daq.setInt('/dev8040/sigouts/0/on', 1), write

        hdawg.seti('sigouts/0/on, 1)

        :node: Node which will be appended to '/device_id/'
        :new_int: New value for integer
        """

        self.daq.setInt(f'/{self.device_id}/{node}', new_int)

    @log_standard_output
    def setd(self, node, new_double):
        """
        Warapper for daq.setDouble commands. For instance, instead of
        daq.setDouble('/dev8040/sigouts/0/range', 0.8), write

        hdawg.setd('sigouts/0/range')

        :node: Node which will be appended to '/device_id/'
        :new_double: New value for double.
        """

        self.daq.setDouble(f'/{self.device_id}/{node}', new_double)

    @log_standard_output
    def setv(self, node, vector):
        """
        Warapper for daq.setVector commands. For instance, instead of
        daq.setVector('/dev8060/awgs/0/waveform/waves/1', vector), write

        hdawg.setd('sigouts/awgs/0/waveform/waves/1', vector)

        :node: Node which will be appended to '/device_id/'
        :new_double: New value for double.
        """

        self.daq.setVector(f'/{self.device_id}/{node}', vector)

    @log_standard_output
    def geti(self, node):
        """
        Warapper for daq.getInt commands. For instance, instead of
        daq.getInt('/dev8040/sigouts/0/busy'), write

        hdawg.geti('sigouts/0/busy')

        :node: Node which will be appended to '/device_id/'
        """

        return self.daq.getInt(f'/{self.device_id}/{node}')

    def set_channel_grouping(self, index):
        """ Specifies channel grouping.

        :index: Integer indicating channel grouping:
            0 : 4x2 with HDAWG8; 2x2 with HDAWG4.
            1 : 2x4 with HDAWG8; 1x4 with HDAWG4.
            2 : 1x8 with HDAWG8.
        """
        self.seti('system/awg/channelgrouping', index)
        time.sleep(2)

    # Functions related to wave outputs:

    def _toggle_output(self, output_indices, target_index):
        """
        Local function enabeling/disabeling wave output.
        """

        # If single integer is given, convert to list.
        output_indices = self._convert_to_list(output_indices)

        for output_index in output_indices:
            if output_index in range(self.num_outputs):
                self.seti(f'sigouts/{output_index}/on', target_index)
                if target_index == 1:
                    self.log.info(f"Enabled wave output {output_index}.")
                elif target_index == 0:
                    self.log.info(f"Disable wave output {output_index}.")
            else:
                self.log.error(
                    f"This device has only {self.num_outputs} channels, \
                        channel index {output_index} is invalid."
                )

    def enable_output(self, output_indices):
        """
        Enables wave output.

        Channel designation uses channel index (0 to 7),
        not channel number (1 to 8).

        :output_index: List or int containing
            integers indicating wave output 0 to 7
        """

        self._toggle_output(output_indices, 1)

    def disable_output(self, output_indices):
        """
        Disables wave output.

        :output_index: List or int containing
            integers indicating wave output 0 to 7
        """
        self._toggle_output(output_indices, 0)

    def set_output_range(self, output_index, output_range):
        """
        Set the output range.

        :output_index: List or int containing integers
            indicating wave output 0 to 7
        :output_range: Double indicating the range of wave output, in Volt.
            All waveforms (ranging from 0 to 1 in value) will be multiplied
            with this value. Possible ranges are:
            0.2, 0.4, 0.6, 0.8, 1, 2, 3, 4, 5 (V)
        """

        allowed_ranges = [0.2, 0.4, 0.6, 0.8, 1, 2, 3, 4, 5]

        if output_index in range(self.num_outputs):
            if output_range in allowed_ranges:

                # Send change range command.
                self.setd(f'sigouts/{output_index}/range', output_range)

                # Wait for HDAWG to be ready, try 100 times before timeout.
                max_tries = 100
                num_tries = 0

                while self.geti(f'sigouts/{output_index}/busy') and num_tries < max_tries:
                    time.sleep(0.2)
                    num_tries += 1

                if num_tries is max_tries:
                    self.log.error(
                        f"Range change timeout after {max_tries} tries."
                    )
                else:
                    self.log.info(
                        f"Changed range of wave output {output_index} to {output_range} V."
                    )
            else:
                self.log.error(
                    f"Range {output_range} is not valid, allowed values for range are {allowed_ranges}"
                )
        else:
            self.log.error(
                f"This device has only {self.num_outputs} channels, channel index {output_index} is invalid."
            )


class AWGModule():
    """ Wrapper class for awgModule"""

    def __init__(self, hdawg_driver, index):
        """ Setup AWG Module

        :index: Which AWG sequencer to be used
            0 - 3 for 4x2 channel grouping
            0 - 1 for 2x4 channel grouping
            0     for 1x8 channel grouping

         :hdawg_driver: Instance of HDAWG_Driver
        """

        self.hd = hdawg_driver
        self.index = index

        # Check if chosen index is allowed for current channel grouping.
        channel_grouping = hdawg_driver.geti('system/awg/channelgrouping')

        if channel_grouping == 0:
            num_awgs = 4
        elif channel_grouping == 1:
            num_awgs = 2
        elif channel_grouping == 2:
            num_awgs = 1

        allowed_indices = range(num_awgs)

        if index not in allowed_indices:
            self.hd.log.error(
                f"Current channel grouping only allows for the following \
                     AWG indices {list(allowed_indices)}"
            )
            return None

        # Create an instance of the AWG Module
        awgModule = hdawg_driver.daq.awgModule()
        awgModule.set('index', index)
        awgModule.set('device', hdawg_driver.device_id)
        awgModule.execute()

        # Disable re-run function
        self.hd.seti(f'awgs/{self.index}/single', 1)

        self.module = awgModule
        self.hd.log.info(f"AWG {self.index}: Module created.")

    def set_sampling_rate(self, sampling_rate):
        """ Set sampling rate of AWG output

        :sampling_rate: String of target sampling rate:
            '2.4 GHz',
            '1.2 GHz',
            '600 MHz',
            '300 MHz',
            '150 MHz',
            '75 MHz',
            '37.5 MHz',
            '18.75 MHz',
            '9.37 MHz',
            '4.68 MHz',
            '2.34 MHz',
            '1.17 MHz',
            '585.93 kHz',
            '292.96 kHz'
        """

        possible_sampling_rates = SAMPLING_RATE_DICT.keys()
        if sampling_rate not in possible_sampling_rates:
            self.hd.log.error(f"AWG {self.index}: Invalid sampling rate '{sampling_rate}', possible choices are {list(possible_sampling_rates)}")
            return

        sampling_rate_index = SAMPLING_RATE_DICT[sampling_rate]

        self.hd.seti(f'awgs/{self.index}/time', sampling_rate_index)
        self.hd.log.info(
            f"AWG {self.index}: Changed sampling rate to {sampling_rate}."
        )

    def start(self):
        """ Start AWG"""
        self.module.set('awg/enable', 1)
        self.hd.log.info(f"AWG {self.index}: Started.")

    def stop(self):
        """ Stop AWG"""
        self.module.set('awg/enable', 0)
        self.hd.log.info(f"AWG {self.index}: Stopped.")

    def compile_upload_sequence(self, sequence):
        """ Compile and upload AWG sequence to AWG Module.

        :sequence: Instance of Sequence class.
        """

        # First check if all values have been replaced in sequence:
        if not sequence.is_ready():
            self.hd.log.error("Sequence is not ready: Not all placeholders have been replaced.")
            return

        self.module.set('compiler/sourcestring', sequence.sequence)
        # Note: when using an AWG program from a source file
        # (and only then), the compiler needs to
        # be started explicitly with awgModule.set('compiler/start', 1)
        while self.module.getInt('compiler/status') == -1:
            time.sleep(0.1)

        if self.module.getInt('compiler/status') == 1:
            # compilation failed, raise an exception
            self.hd.log.warn(self.module.getString('compiler/statusstring'))

        if self.module.getInt('compiler/status') == 0:
            self.hd.log.info(
                "Compilation successful with no warnings, will upload the program to the instrument."
            )
        if self.module.getInt('compiler/status') == 2:
            self.hd.log.warn(
                "Compilation successful with warnings, will upload the program to the instrument."
            )
            self.hd.log.warn(
                f"Compiler warning: {self.module.getString('compiler/statusstring')}"
            )

        # Wait for the waveform upload to finish
        time.sleep(0.2)
        i = 0
        while (self.module.getDouble('progress') < 1.0) and (self.module.getInt('elf/status') != 1):
            self.hd.log.info("{} progress: {:.2f}".format(i, self.module.getDouble('progress')))
            time.sleep(0.2)
            i += 1
        self.hd.log.info(
            "{} progress: {:.2f}".format(i, self.module.getDouble('progress'))
        )
        if self.module.getInt('elf/status') == 0:
            self.hd.log.info("Upload to the instrument successful.")
        if self.module.getInt('elf/status') == 1:
            self.hd.log.warning("Upload to the instrument failed.")

    def dyn_waveform_upload(self, index, waveform1, waveform2=None):
        """ Dynamically upload a numpy array into HDAWG Memory

        This will overwrite the allocated waveform memory of a waveform
        defined in the sequence. The index designates which waveform to
        overwrite:
        Let N be the total number of waveforms and M>0 be the number of
        waveforms defined from CSV files. Then the index
        of the waveform to be replaced is defined as following:
        - 0,...,M-1 for all waveforms defined from CSV file
            alphabetically ordered by filename,
        - M,...,N-1 in the order that the waveforms are
            defined in the sequencer program.
        For the case of M=0, the index is defined as:
        - 0,...,N-1 in the order that the waveforms are
            defined in the sequencer program.

        :waveform1: np.array containing waveform
        :waveform2: np.array containing waveform of second waveform is dynamic
            waveform upload is used for playback of two waveforms at two channels,
            as represented in the .seqc command `playWave(waveform1, waveform2)
        :index: Index of waveform to be overwritten as defined above
        """

        waveform_native = zhinst.utils.convert_awg_waveform(
            waveform1,
            waveform2
        )

        awg_index = self.module.get('index')['index'][0]
        self.hd.setv(
            f'awgs/{awg_index}/waveform/waves/{index}', waveform_native
        )


class Sequence():
    """ Helper class containing .seqc sequences and helper functions

    """

    def replace_placeholder(self, placeholder, value):
        """ Replace a placeholder by some value

        :placeholder: Placeholder string to be replaced.
        :value: Value to which the placeholder string need to be set.
        """
        self.sequence = self.sequence.replace(f"_{placeholder}_", str(value))
        self.unresolved_placeholders.remove(placeholder)

    def replace_waveform(self, placeholder, waveform):
        """ Replace a placeholder by a waveform

        :placeholder: Placeholder string to be replaced.
        :waveform: Numpy array designating the waveform.
        """
        waveform = 'vect(' + ','.join([str(x) for x in waveform]) + ')'
        self.sequence = self.sequence.replace(f"_{placeholder}_", waveform)
        self.unresolved_placeholders.remove(placeholder)

    def is_ready(self):
        """ Return True if all placeholders have been replaced"""
        return len(self.unresolved_placeholders) == 0

    def __init__(self, hdawg_driver, sequence, placeholders=None):
        """ Initialize sequence with string

        :hdawg_driver: Instance of HDAWG_Driver
        :sequence: A string which contains sequence instructions,
            as defined in the ZI language seqc. This string can contain placeholders
            indicated by '_c_', where c is the name of the placeholder.
        :placeholders: A list of placeholders which need to be replaced
            before compilation of sequence. If for example '_c_' is included
            in the sequence string, 'c' is the name of the placeholder to be
            passed in this argument.
        """

        # Store reference to HDAWG_Driver to use logging function.
        self.hd = hdawg_driver

        # Some sanity checks.
        for placeholder in placeholders:
            if f"_{placeholder}_" not in sequence:
                error_msg = f"The placeholder _{placeholder}_ cannot \
                    be found in the sequence."
                hdawg_driver.log.error(error_msg)
                raise Exception(error_msg)

        # Store sequence and placeholders.
        self.sequence = textwrap.dedent(sequence)
        self.placeholders = placeholders
        # Keeps track of which placeholders has not been replaced yet.
        self.unresolved_placeholders = copy.deepcopy(placeholders)



import pickle

from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Service(ServiceBase):

    def exposed_start_counting(self, bin_width=1000000000, n_bins=10000):
        return self._module.start_ctr(
            bin_width=bin_width,
            n_bins=n_bins
        )

    def exposed_clear_counter(self):
        return self._module.clear_ctr()

    def exposed_get_counts(self):
        res_pickle = self._module.get_counts()
        return pickle.dumps(res_pickle)

    def exposed_get_x_axis(self):
        res_pickle = self._module.get_x_axis()
        return pickle.dumps(res_pickle)

    def exposed_set_channels(self, ch_list=[1]):
        return self._module.set_ch_assignment(
            ch_list=ch_list
        )


class Client(ClientBase):

    def start_counting(self, bin_width=1000000000, n_bins=10000):
        return self._service.exposed_start_counting(
            bin_width=bin_width,
            n_bins=n_bins
        )

    def clear_counter(self):
        return self._service.exposed_clear_counter()

    def get_counts(self):
        res_pickle = self._service.exposed_get_counts()
        return pickle.loads(res_pickle)

    def get_x_axis(self):
        res_pickle = self._service.exposed_get_x_axis()
        return pickle.loads(res_pickle)

    def set_channels(self, ch_list=[1]):
        return self._service.exposed_set_channels(
            ch_list=ch_list
        )




""" Constant definitions and function prototypes from
National Instruments High-Speed Digital I/O
API header file niHSDIO.h
"""


import ctypes

# #if defined(LONG_MAX) && (LONG_MAX > 0x7FFFFFFFL)
# typedef unsigned int        ViUInt32;
# typedef _VI_SIGNED int      ViInt32;
# #else
# typedef unsigned long       ViUInt32;
# typedef _VI_SIGNED long     ViInt32;
# #endif


class NITypes:
    ViUInt16 = ctypes.c_ushort  # for ViPUInt16 - use ct.byref()
    ViUInt32 = ctypes.c_uint32
    ViInt32 = ctypes.c_int32

    ViReal64 = ctypes.c_double

    ViBoolean = ViUInt16

    ViChar = ctypes.c_char
    ViPChar = ctypes.c_char_p
    ViString = ViPChar
    ViConstString = ViPChar

    ViAttr = ViUInt32
    ViRsrc = ViString
    ViObject = ViUInt32
    ViSession = ViObject
    ViStatus = ViInt32


class NIConst:
    # ========== Symbolic constants =================
    VI_TRUE = NITypes.ViBoolean(1)
    VI_FALSE = NITypes.ViBoolean(0)
    VI_NULL = NITypes.ViUInt32(0)
    VI_SUCCESS = ctypes.c_long(0)

    NIHSDIO_VAL_ON_BOARD_CLOCK_STR = NITypes.ViConstString(b"OnBoardClock")

    # Attributes
    IVI_ATTR_BASE = 1000000
    IVI_SPECIFIC_PUBLIC_ATTR_BASE = IVI_ATTR_BASE + 150000

    # /***************************************************************************************************
    #  *------------------------------------------ Attribute Defines ------------------------------------*
    #  ***************************************************************************************************/

    # /* IVI Inherent Instrument Attributes */

    # /* Advanced Session Information */

    # /********************************** Instrument-Specific Attributes *********************************/

    # /* Base attributes */
    NIHSDIO_ATTR_DYNAMIC_CHANNELS = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 2)  # /* ViString */
    NIHSDIO_ATTR_STATIC_CHANNELS = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 3)  # /* ViString */

    # /* Device attributes */
    NIHSDIO_ATTR_TOTAL_GENERATION_MEMORY_SIZE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 74)  # / * ViInt32 * /

    # /* Voltage attributes */
    NIHSDIO_ATTR_DATA_VOLTAGE_LOW_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 6)   # ViReal64, channel-based
    NIHSDIO_ATTR_DATA_VOLTAGE_HIGH_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 7)  # ViReal64, channel-based
    NIHSDIO_ATTR_DATA_TERMINATION_VOLTAGE_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 161)  # ViReal64, channel-based
    NIHSDIO_ATTR_DATA_VOLTAGE_RANGE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 163)  # /* ViInt32 */
    NIHSDIO_ATTR_TRIGGER_VOLTAGE_LOW_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 8)   # /* ViReal64 */
    NIHSDIO_ATTR_TRIGGER_VOLTAGE_HIGH_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 9)  # /* ViReal64 */
    NIHSDIO_ATTR_EVENT_VOLTAGE_LOW_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 79)    # /* ViReal64 */
    NIHSDIO_ATTR_EVENT_VOLTAGE_HIGH_LEVEL = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 80)   # /* ViReal64 */

    # /* Channel attributes */

    # /* Clocking attributes */

    # /* Triggers attributes */

    # /* Events attributes */

    # /* Dynamic output attributes */
    NIHSDIO_ATTR_GENERATION_MODE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 25)       # /* ViInt32 */
    NIHSDIO_ATTR_REPEAT_MODE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 26)           # /* ViInt32 */
    NIHSDIO_ATTR_REPEAT_COUNT = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 71)          # /* ViInt32 */
    NIHSDIO_ATTR_WAVEFORM_TO_GENERATE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 27)  # /* ViString */
    NIHSDIO_ATTR_SCRIPT_TO_GENERATE = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 28)    # /* ViString */

    # /* Dynamic input attributes */

    # /* Data Position Delay */

    # /* Data Width */
    NIHSDIO_ATTR_DATA_WIDTH = NITypes.ViAttr(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 108)   # / * ViInt32 * /

    # /* Data Rate Multiplier */

    # /* Supported Data States */

    # /* Synchronization attributes */

    # /* Hardware compare */

    # /* Streaming attributes */

    # /* DirectDMA attributes */

    # /* Bandwidth Governing attributes */

    # /* Power Management attributes */

    # /***************************************************************************************************
    #  *----------------------------------- Attribute Value Defines -------------------------------------*
    #  ***************************************************************************************************/

    # /************************** Instrument specific attribute value definitions ************************/

    # /* Data interpretation */

    # /* Drive type */

    # /* Logic family */

    # /* Data Voltage Range */

    # /* Tristate Mode */

    # /* Digital edge */

    # /* Pattern Match Values */

    # /* Trigger digital pattern trigger when */

    # /* Event polarity */

    # /* Generation mode */
    NIHSDIO_VAL_WAVEFORM = NITypes.ViInt32(14)
    NIHSDIO_VAL_SCRIPTED = NITypes.ViInt32(15)

    # /* Repeat mode */
    NIHSDIO_VAL_FINITE = NITypes.ViInt32(16)
    NIHSDIO_VAL_CONTINUOUS = NITypes.ViInt32(17)

    # /* Initial/Idle State */

    # /* Trigger Types */

    # /* Level Trigger Values */

    # /* Signal Values */

    # /* Terminal Configuration Values */

    # /****** String enum values for signal names *******/
    # #define NIHSDIO_VAL_SCRIPT_TRIGGER0                      "scriptTrigger0"
    # #define NIHSDIO_VAL_SCRIPT_TRIGGER1                      "scriptTrigger1"
    # #define NIHSDIO_VAL_SCRIPT_TRIGGER2                      "scriptTrigger2"
    # #define NIHSDIO_VAL_SCRIPT_TRIGGER3                      "scriptTrigger3"
    # #define NIHSDIO_VAL_MARKER_EVENT0                        "marker0"
    # #define NIHSDIO_VAL_MARKER_EVENT1                        "marker1"
    # #define NIHSDIO_VAL_MARKER_EVENT2                        "marker2"
    # #define NIHSDIO_VAL_MARKER_EVENT3                        "marker3"





    # NIHSDIO_ATTR_DATA_WIDTH = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 108)
    NIHSDIO_ATTR_SAMPLE_CLOCK_SOURCE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 13)  # output type: ViString
    NIHSDIO_ATTR_SAMPLE_CLOCK_RATE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 14)
    NIHSDIO_ATTR_SERIAL_NUMBER = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 96)

    # Generation mode
    # NIHSDIO_VAL_WAVEFORM = NITypes.ViInt32(14)
    # NIHSDIO_VAL_SCRIPTED = NITypes.ViInt32(15)
    # NIHSDIO_ATTR_GENERATION_MODE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 25)  # output type: ViInt32

    # Repeat mode and count (use only for Waveform mode)
    # NIHSDIO_VAL_FINITE = NITypes.ViInt32(16)
    # NIHSDIO_VAL_CONTINUOUS = NITypes.ViInt32(17)
    # NIHSDIO_ATTR_REPEAT_MODE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 26)  # output type: ViInt32
    # NIHSDIO_ATTR_REPEAT_COUNT = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 71)  # output type: ViInt32

    # Waveform/Script to generate
    # NIHSDIO_ATTR_WAVEFORM_TO_GENERATE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 27)  # output type: ViString
    # NIHSDIO_ATTR_SCRIPT_TO_GENERATE = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 28)  # output type: ViString

    # Logic Level Voltage
    NIHSDIO_VAL_5_0V_LOGIC = NITypes.ViInt32(5)
    NIHSDIO_VAL_3_3V_LOGIC = NITypes.ViInt32(6)
    NIHSDIO_VAL_2_5V_LOGIC = NITypes.ViInt32(7)
    NIHSDIO_VAL_1_8V_LOGIC = NITypes.ViInt32(8)
    NIHSDIO_VAL_1_5V_LOGIC = NITypes.ViInt32(80)
    NIHSDIO_VAL_1_2V_LOGIC = NITypes.ViInt32(81)

    # NIHSDIO_ATTR_DATA_VOLTAGE_LOW_LEVEL = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 6)
    # NIHSDIO_ATTR_DATA_VOLTAGE_HIGH_LEVEL = NITypes.ViInt32(IVI_SPECIFIC_PUBLIC_ATTR_BASE + 7)


# ========= Function prototypes =================

def build_c_func_prototypes(ni_hsdio_dll_instance):
    ni_hsdio_dll_instance.niHSDIO_error_message.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_InitGenerationSession.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_AssignDynamicChannels.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureSampleClock.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureDataVoltageLogicFamily.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureDataVoltageCustomLevels.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_GetAttributeViInt32.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_GetAttributeViReal64.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_GetAttributeViString.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_WriteNamedWaveformU32.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_DeleteNamedWaveform.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_WriteScript.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_ConfigureGenerationMode.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureGenerationRepeat.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureWaveformToGenerate.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_ConfigureScriptToGenerate.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_ConfigureDigitalEdgeStartTrigger.restype = NITypes.ViStatus

    ni_hsdio_dll_instance.niHSDIO_Initiate.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_Abort.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_close.restype = NITypes.ViStatus
    ni_hsdio_dll_instance.niHSDIO_reset.restype = NITypes.ViStatus



# """ Broken due to dependency on pulseblock. Should be fixed once pulseblock is integrated"""

# from pylabnet.utils.logging.logger import LogHandler
# from pylabnet.hardware.interface.simple_p_gen import PGenError
# import pulseblock.pulse as po
# from pulseblock.pb_sample import pb_sample
# from pulseblock.pb_zip import pb_zip

# import os
# import time
# import visa
# import numpy as np
# from ftplib import FTP
# import copy

# from pylabnet.network.core.service_base import ServiceBase
# from pylabnet.network.core.client_base import ClientBase


# class Driver:
#     """ A hardware module for the Tektronix AWG7000 series for generating
#     waveforms and sequences thereof.

#     Adapted from github.com/Ulm-IQO/qudi.
#     """

#     def __init__(self,
#                  visa_adr_str, ftp_ip_str,
#                  ftp_username=None, ftp_pswrd=None,
#                  local_wfm_dir='user\\local_wfm_dir',
#                  remote_wfm_dir='remote_wfm_dir',
#                  visa_timeout=20,
#                  logger=None):

#         self.log = LogHandler(logger=logger)

#         # Declaration of internal variables -----------------------------------

#         self._awg = None  # This variable will hold a reference to the awg visa resource

#         self.option_list = list()  # will hold the encoded installed options available on _awg
#         self.model = None
#         self._local_wfm_dir = None
#         self._ftp_ip_str = None
#         self._ftp_username = None
#         self._ftp_pswrd = None
#         self._written_seqs = []  # Helper variable since written sequences can not be queried
#         self._loaded_seqs = []  # Helper variable since a loaded sequence can not be queried :(

#         # VISA connection -----------------------------------------------------

#         self._rm = visa.ResourceManager()
#         try:
#             self._awg = self._rm.open_resource(visa_adr_str)
#             # set timeout (in ms)
#             self._awg.timeout = visa_timeout * 1000
#             # Reset hardware
#             self.reset()
#         except:
#             self._awg = None

#             msg_str = 'VISA connection to address {} failed'.format(visa_adr_str)

#             self.log.exception(msg_str=msg_str)
#             raise PGenError(msg_str)

#         # FTP connection ------------------------------------------------------

#         # FTP (File Transfer Protocol) is used to transfer large waveform
#         # from PC's hard drive to AWG's hard drive:
#         #   - PulseBlock is sampled and the binary is stored in local_wfm_dir
#         #   - FTP transfers file to remote_wfm_dir on AWG's hard drive
#         #   - AWG loads the wfm into fast memory

#         self._ftp_ip_str = ftp_ip_str
#         self._ftp_username = ftp_username
#         self._ftp_pswrd = ftp_pswrd
#         self._remote_wfm_dir = remote_wfm_dir

#         # Test FTP connection and set current working dir to remote_wfm_dir
#         try:
#             with FTP(ftp_ip_str) as ftp:
#                 ftp.login(user=self._ftp_username, passwd=self._ftp_pswrd)
#                 ftp.cwd(self._remote_wfm_dir)
#                 self.log.debug('FTP connection test. Working dir: {0}'.format(ftp.pwd()))
#         except:
#             msg_str = 'Attempt to establish FTP connection failed. \n' \
#                       'The following params where used: \n' \
#                       '     ftp_ip_str = {} \n' \
#                       '     ftp_username = {} \n' \
#                       '     ftp_pswrd = {} \n' \
#                       '     remote_wfm_dir = {}' \
#                       ''.format(ftp_ip_str, ftp_username, ftp_pswrd, remote_wfm_dir)

#             self.log.exception(msg_str=msg_str)
#             raise PGenError(msg_str)

#         # Create local_wfm_dir
#         if local_wfm_dir == 'user\\local_wfm_dir':
#             local_wfm_dir = os.path.expanduser('~\\local_wfm_dir')

#         if not os.path.exists(local_wfm_dir):
#             os.makedirs(os.path.abspath(local_wfm_dir))

#         self._local_wfm_dir = local_wfm_dir

#         # Determine device params ---------------------------------------------

#         # Get model and option list
#         mfg, self.model, ser, fw_ver = self.query('*IDN?').split(',')
#         self.option_list = self.query('*OPT?').split(',')
#         # Options of AWG7000 series:
#         #              Option 01: Memory expansion to 64,8 MSamples (Million points)
#         #              Option 06: Interleave and extended analog output bandwidth
#         #              Option 08: Fast sequence switching
#         #              Option 09: Subsequence and Table Jump

#         # Log confirmation info message ---------------------------------------
#         self.log.info(
#             'Found {} {} Serial: {} FW: {} \n'
#             'Installed options: {}'.format(
#                 mfg, self.model, ser, fw_ver, self.option_list
#             )
#         )

#     def close(self):
#         """ Deinitialisation performed during deactivation of the module.
#         """

#         # Closes the connection to the AWG
#         try:
#             self._awg.close()
#         except:
#             self.log.debug('Closing AWG connection using pyvisa failed.')

#         self.log.info('Closed connection to AWG')

#         return 0

#     # -------------------------------------------------------------------------
#     # Basic commands and Hardware settings
#     # -------------------------------------------------------------------------

#     # Basic commands

#     def write(self, cmd_str):
#         """ Sends a command string to the device.

#         Should only be used for non-responding (set-type) commands!
#         If the command is supposed to produce some response message,
#         use query() method.

#         @param string cmd_str: string containing the command
#         @return int: error code (0:OK, -1:error)
#         """

#         self._awg.write('*WAI')
#         bytes_written, status_code = self._awg.write(cmd_str)
#         # self._awg.write('*WAI')

#         # Block Python process until the operation is complete
#         # TODO: implement more robust blocking mechanism:
#         # TODO: to avoid timeout for very long operations
#         self._awg.query('*OPC?')

#         self.raise_errors()

#         return status_code

#     def query(self, question):
#         """ Query the device:
#             write request, read response, return formatted answer.

#         @param string question: string containing the command

#         @return string: the answer of the device to the 'question' in a string
#         """

#         try:
#             answer = self._awg.query(question)
#             answer = answer.strip()
#             answer = answer.rstrip('\n')
#             answer = answer.rstrip()
#             answer = answer.strip('"')
#             return answer

#         except Exception as exc_obj:
#             self.log.exception(msg_str='Exception in query()')

#             # If the reason for timeout is an incorrect command,
#             # raise_errors() should produce an exception.
#             self.raise_errors()
#             # This call is necessary to clean-up the error register
#             # such that subsequent commands do not break due to this
#             # error message still in the register.

#             # Else, some other error was produced.
#             # Just re-raise the original exception
#             raise exc_obj

#     def reset(self):
#         """ Reset the device.

#         @return int: error code (0:OK, -1:error)
#         """

#         self.write('*RST')

#         # Clear status register
#         self.write('*CLS')

#         return 0

#     def get_status(self):
#         """ Retrieves the status of the pulsing hardware

#         -1: 'Failed Request or Communication'
#          0: 'Device has stopped, but can receive commands'
#          1: 'Device is waiting for trigger'
#          2: 'Device is active and running'

#         @return (int, dict): inter value of the current status with the
#                              corresponding dictionary containing status
#                              description for all the possible status variables
#                              of the pulse generator hardware
#         """

#         current_status = -1 if self._awg is None else int(self.query('AWGC:RST?'))
#         return current_status

#     def start(self):
#         """ Switches the pulsing device on.

#         @return int: error code (0:OK, -1:error, higher number corresponds to
#                                  current status of the device. Check then the
#                                  class variable status_dic.)
#         """

#         # Switch outputs on
#         # (here it is assumed that all active analog channels have
#         # a waveform assigned, such that all active channels are switched on)
#         for ch in self._get_all_anlg_chs():
#             ch_num = int(ch.rsplit('_ch', 1)[1])
#             self.write(
#                 'OUTPUT{}:STATE ON'.format(ch_num)
#             )

#         # Start generator
#         self.write('AWGC:RUN')

#         return 0

#     def stop(self):
#         """ Switches the pulsing device off.

#         @return int: error code (0:OK, -1:error, higher number corresponds to
#                                  current status of the device. Check then the
#                                  class variable status_dic.)
#         """

#         self.write('AWGC:STOP')

#         for ch in self._get_all_anlg_chs():
#             ch_num = int(ch.rsplit('_ch', 1)[1])
#             self.write(
#                 'OUTPUT{0:d}:STATE OFF'.format(ch_num)
#             )

#         return 0

#     def raise_errors(self):
#         """Get all errors from the device and log them.

#         :return: (int) 0 is there is no errors.
#                  PGenError is produced in the case of error.
#         """

#         # Get all errors
#         read_next_err = True
#         has_error = False

#         msg_str_list = []

#         while read_next_err:
#             err = self.query('SYST:ERR?').split(',')

#             if int(err[0]) == 0:
#                 read_next_err = False
#             else:
#                 has_error = True
#                 msg_str_list.append(
#                     '{} {}'.format(err[0], err[1])
#                 )

#         # Log/raise if there are some errors
#         if has_error:

#             # Combine all error messages into a single string
#             # with one error per line
#             msg_str = '\n'.join(msg_str_list)

#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return 0

#     # Hardware settings

#     def get_mode(self):
#         """
#         Returns AWG run mode according to the following list:
#             continuous - 'C'
#             triggered  - 'T'
#             gated      - 'G'
#             sequence   - 'S'
#         @return str: 'C', 'T', 'G', 'S'.
#         """

#         run_mode_dict = {'CONT': 'C', 'TRIG': 'T', 'GAT': 'G', 'SEQ': 'S'}

#         answer = self.query('AWGC:RMOD?')
#         if answer in run_mode_dict:
#             return run_mode_dict[answer]
#         else:
#             msg_str = 'get_mode(): returned answer {} is not on the list of known run modes' \
#                       ''.format(answer)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#     def set_mode(self, mode_str):
#         """Change the run mode of the AWG7000 series.

#         :param str mode_str: Options for mode (case-insensitive):
#                             continuous - 'C'
#                             triggered  - 'T'
#                             gated      - 'G'
#                             sequence   - 'S'
#         """

#         look_up_dict = {
#             'C': 'CONT',
#             'T': 'TRIG',
#             'G': 'GAT',
#             'E': 'ENH',
#             'S': 'SEQ'
#         }
#         self.write(
#             'AWGC:RMOD {0!s}'.format(
#                 look_up_dict[mode_str.upper()]
#             )
#         )

#         return self.get_mode()

#     def get_samp_rate(self):
#         """ Get the sample rate of the pulse generator hardware

#         :return float: The current sample rate of the device (in Hz)
#         """

#         return float(
#             self.query('SOUR1:FREQ?')
#         )

#     def set_samp_rate(self, samp_rate):
#         """ Set the sample rate of the pulse generator hardware.

#         :param float samp_rate: The sampling rate to be set (in Hz)

#         :return float: the sample rate returned from the device (in Hz).

#         Note: After setting the sampling rate of the device, use the actually set return value for
#               further processing.
#         """

#         self.write('SOUR1:FREQ {} Hz'.format(samp_rate))

#         return self.get_samp_rate()

#     def get_analog_level(self):
#         """ Retrieve the analog amplitude and offset of the provided channels.

#         :return: nested dictionary:
#                 dict(
#                     'a_ch1'={'amp_pp': amp_1, 'offset': offset_1},
#                     'a_ch2'={'amp_pp': amp_2, 'offset': offset_2}
#                 )
#                 Outer keys (str)- the channel name (i.e. 'a_ch1' and 'a_ch2')
#                 Inner keys (str) - 'amp_pp' and 'offset'
#                 Items - the values for those channels.
#                 Amplitude is always denoted in Volt-peak-to-peak and Offset in volts.
#         """

#         level_dict = dict()

#         # Depending on included options, offset may not be available
#         no_offset = ('02' in self.option_list) or ('06' in self.option_list)

#         for ch_name in self._get_all_anlg_chs():
#             ch_num = int(ch_name.rsplit('_ch', 1)[1])
#             level_dict[ch_name] = dict()

#             # Amplitude-PP
#             level_dict[ch_name]['amp_pp'] = float(
#                 self.query('SOUR{0:d}:VOLT:AMPL?'.format(ch_num))
#             )

#             # Offset
#             level_dict[ch_name]['offset'] = 0.0 if no_offset else float(
#                 self.query('SOUR{0:d}:VOLT:OFFS?'.format(ch_num))
#             )

#         return level_dict

#     def set_analog_level(self, level_dict):
#         """ Set amplitude-pp and/or offset.

#         :param level_dict: nested dictionary:
#                 dict(
#                     'a_ch1'={'amp_pp': amp_1, 'offset': offset_1},
#                     'a_ch2'={'amp_pp': amp_2, 'offset': offset_2}
#                 )
#                 Outer keys (str)- the channel number (i.e. 'a_ch1' and 'a_ch2')
#                 Inner keys (str) - 'amp_pp' and 'offset'
#                 Items - the values for those channels [in Volts].

#                 If some channels/parameters are not mentioned, they are not changed.

#         :return: nested dict of the same structure as level_dict
#         """

#         for ch_name in level_dict.keys():
#             ch_num = int(ch_name.rsplit('_ch', 1)[1])

#             # set amplitude-pp
#             if 'amp_pp' in level_dict[ch_name].keys():
#                 self.write(
#                     'SOUR{}:VOLT:AMPL {}'.format(
#                         ch_num,
#                         level_dict[ch_name]['amp_pp']
#                     )
#                 )

#             # set offset
#             if 'offset' in level_dict[ch_name].keys():
#                 self.write(
#                     'SOUR{}:VOLT:OFFSET {}'.format(
#                         ch_num,
#                         level_dict[ch_name]['offset']
#                     )
#                 )

#         return self.get_analog_level()

#     def get_digital_level(self):
#         """ Retrieve the digital low and high level of all marker channels.

#         :return dict:
#                 {
#                     'd_ch1': {'low': 0.1, 'high': 1.2},
#                     'd_ch2': {'low': 0.0, 'high': 1.0},
#                     'd_ch3': {'low': 0.0, 'high': 1.0},
#                     'd_ch4': {'low': 0.0, 'high': 1.0}

#                 }
#         """

#         level_dict = dict()

#         for ch_name in self._get_all_dgtl_chs():
#             level_dict[ch_name] = dict()
#             d_ch_num = int(ch_name.rsplit('_ch', 1)[1])

#             # Convert into 'Analog channel + marker number'
#             a_ch_num = (d_ch_num + 1) // 2
#             d_ch_num -= 2 * (a_ch_num - 1)

#             # Low
#             level_dict[ch_name]['low'] = float(
#                 self.query(
#                     'SOUR{}:MARK{}:VOLT:LOW?'.format(a_ch_num, d_ch_num)
#                 )
#             )
#             # High
#             level_dict[ch_name]['high'] = float(
#                 self.query(
#                     'SOUR{}:MARK{}:VOLT:HIGH?'.format(a_ch_num, d_ch_num)
#                 )
#             )

#         return level_dict

#     def set_digital_level(self, level_dict):
#         """ Set low/high value of marker channels.

#         :param level_dict: nested dict of the following structure:
#             {
#                     'd_ch1': {'low': 0.1, 'high': 1.2},
#                     'd_ch2': {'low': 0.0, 'high': 1.0},
#                     'd_ch3': {'low': 0.0, 'high': 1.0},
#                     'd_ch4': {'low': 0.0, 'high': 1.0}

#             }
#         If some channels/markers/values are not mentioned, corresponding values
#         are not changed.

#         :return dict: the actual levels.
#         """

#         # Determine the requested final state:
#         #   get current state
#         req_dict = self.get_digital_level()
#         #   update it with the new values from level_dict
#         for d_ch_name in level_dict.keys():
#             try:
#                 if 'low' in level_dict[d_ch_name]:
#                     req_dict[d_ch_name]['low'] = level_dict[d_ch_name]['low']

#                 if 'high' in level_dict[d_ch_name]:
#                     req_dict[d_ch_name]['high'] = level_dict[d_ch_name]['high']

#             except KeyError:
#                 msg_str = 'set_digital_level(): channel {} is not available. \n' \
#                           'Currently available digital channels: \n' \
#                           '     {}'.format(d_ch_name, self._get_all_dgtl_chs())

#                 self.log.exception(msg_str=msg_str)
#                 raise PGenError(msg_str)

#         # Set all values
#         for d_ch_name in req_dict.keys():
#             d_ch_num = int(d_ch_name.rsplit('_ch', 1)[1])

#             # Convert into 'Analog channel + marker number'
#             a_ch_num = (d_ch_num + 1) // 2
#             d_ch_num -= 2 * (a_ch_num - 1)

#             # Low
#             self.write(
#                 'SOUR{}:MARK{}:VOLT:LOW {}'.format(
#                     a_ch_num,
#                     d_ch_num,
#                     req_dict[d_ch_name]['low']
#                 )
#             )

#             # High
#             self.write(
#                 'SOUR{}:MARK{}:VOLT:HIGH {}'.format(
#                     a_ch_num,
#                     d_ch_num,
#                     req_dict[d_ch_name]['high']
#                 )
#             )

#         # Check if the operation was successful
#         #   - check for explicit error messages was already performed in
#         #     self.write()

#         #   - check if the actual set values coincide with requested:
#         #     if the requested values do not satisfy hardware restrictions,
#         #     device will try to adjust previously set value [low] when setting
#         #     the next one [high]. This may mask the error: no explicit exception
#         #     will be produced, but the actual values will be different from expected.
#         actual_dict = self.get_digital_level()
#         if req_dict != actual_dict:
#             msg_str = 'set_digital_level(): the actual set value is different from expected: \n' \
#                       'requested = {} \n' \
#                       '   actual = {} \n' \
#                       'Most probably the requested values do not satisfy hardware restrictions ' \
#                       'on min/max amplitude [high-low] or overall voltage window.' \
#                       ''.format(req_dict, actual_dict)

#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return self.get_digital_level()

#     def get_interleave(self):
#         """ Check whether Interleave is ON or OFF in AWG.

#         :return bool: True: ON, False: OFF

#         Will always return False for pulse generator hardware without interleave.
#         """

#         if self._has_interleave():
#             return bool(int(
#                 self.query('AWGC:INT:STAT?')
#             ))
#         else:
#             return False

#     def set_interleave(self, state):
#         """ Turns the interleave of an AWG on or off.

#         :param bool state: The state the interleave should be set to
#                            (True: ON, False: OFF)

#         :return bool: actual interleave status (True: ON, False: OFF)
#         """

#         if self._has_interleave():
#             self.write(
#                 'AWGC:INT:STAT {}'.format(int(state))
#             )

#         # Enable all markers for all active analog channels
#         # (set DAC resolution to 8)
#         # FIXME: implement more generic (devices with 14-bit DAC)

#         for a_ch_name in self._get_all_anlg_chs():
#             a_ch_num = int(a_ch_name.rsplit('_ch', 1)[1])
#             self.write(
#                 'SOUR{}:DAC:RES {}'.format(
#                     a_ch_num,
#                     8
#                 )
#             )

#         return self.get_interleave()

#     def set_lowpass_filter(self, ch_num, cutoff_freq):
#         """ Set a lowpass filter to the analog channels of the AWG.

#         :param int ch_num: To which channel to apply, either 1 or 2.
#         :param cutoff_freq: Cutoff Frequency of the lowpass filter in Hz.

#         :return: (int) success code 0. Exception is produced in the case of error.
#         """

#         # Low-pass filter is not available on instruments with options 02 and 06
#         if '02' in self.option_list or '06' in self.option_list:
#             msg_str = 'set_lowpass_filter(): low-pass filter is not available for options [02] and [06].'
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         self.write(
#             'OUTPUT{}:FILTER:LPASS:FREQUENCY {}'
#             ''.format(ch_num, cutoff_freq)
#         )

#         return 0

#     #  Technical methods

#     def _get_all_chs(self):
#         """Helper method to return a sorted list of all technically available
#         channel names (e.g. ['a_ch1', 'a_ch2', 'd_ch1', 'd_ch2', 'd_ch3', 'd_ch4'])

#         :return list: Sorted list of channel names
#         """

#         # Determine number of available channels based on model
#         if self.model == 'AWG7122C':
#             available_chs = ['a_ch1', 'd_ch1', 'd_ch2']

#             if not self.get_interleave():
#                 available_chs.extend(['a_ch2', 'd_ch3', 'd_ch4'])

#         else:
#             msg_str = 'Failed to determine number of available channels: \n' \
#                       'number of channels is not specified for "{}" model. ' \
#                       'Add this model to _get_all_chs()' \
#                       ''.format(self.model)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return sorted(available_chs)

#     def _get_all_anlg_chs(self):
#         """
#         Helper method to return a sorted list of all technically available analog channel
#         descriptors (e.g. ['a_ch1', 'a_ch2'])

#         :return list: Sorted list of analog channels
#         """

#         return sorted(
#             ch for ch in self._get_all_chs() if ch.startswith('a')
#         )

#     def _get_all_dgtl_chs(self):
#         """
#         Helper method to return a sorted list of all technically available digital channel
#         descriptors (e.g. ['d_ch1', 'd_ch2'])

#         :return list: Sorted list of digital channels
#         """

#         return sorted(
#             ch for ch in self._get_all_chs() if ch.startswith('d')
#         )

#     @staticmethod
#     def _is_digital(ch_name):
#         if 'd' in ch_name:
#             return True
#         else:
#             return False

#     def _zeroing_enabled(self):
#         """
#         Checks if the zeroing option is enabled. Only available on devices with option '06'.

#         @return bool: True: enabled, False: disabled
#         """
#         if self._has_interleave():
#             return bool(int(self.query('AWGC:INT:ZER?')))
#         return False

#     def _has_interleave(self):
#         """ Check if the device has the interleave option installed

#             :return bool: device has interleave option
#         """

#         return '06' in self.option_list

#     # -------------------------------------------------------------------------
#     # Waveform
#     # -------------------------------------------------------------------------

#     def fill_unused_chs(self, pb_obj):
#         """Fill all unused channels with default values:
#             DFalse() for marker channels
#             DConst(val=0.0) for analog channels
#         (channel mapping from user-friendly names [like 'AOM' and 'Ctr']
#         to device-native names [like 'a_ch1', 'd_ch2'] has to be performed
#         before calling this function)

#         :param pb_obj: PulseBlock object
#         :return: new PulseBlock object with all unused channels
#                  filled with default values.
#         """

#         pb_obj = copy.deepcopy(pb_obj)

#         pb_ch_set = set(pb_obj.p_dict.keys()) | set(pb_obj.dflt_dict.keys())
#         avail_ch_set = set(self._get_all_chs())

#         # Sanity check: all requested channels are available
#         if not pb_ch_set.issubset(avail_ch_set):
#             msg_str = 'write_wfm(): the following channels of the PulseBlock are not available: \n' \
#                       '{} \n' \
#                       ''.format(sorted(list(pb_ch_set - avail_ch_set)))

#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         # All available channels have to have a waveform assigned.
#         # If some channels are not used in the pb_obj,
#         # fill them with default value pulses
#         unused_ch_set = avail_ch_set - pb_ch_set
#         for unused_ch_name in unused_ch_set:
#             if self._is_digital(unused_ch_name):
#                 pb_obj.dflt_dict[unused_ch_name] = po.DFalse()
#             else:
#                 pb_obj.dflt_dict[unused_ch_name] = po.DConst(val=0.0)

#         return pb_obj

#     def write_wfm(self, pb_obj, len_adj=True, strict_hrdw_seq=False):
#         """Write plain waveform to AWG memory

#         :param pb_obj: PulseBlock object
#         :param len_adj: (bool) [argument of pb_sample()]
#             Add trailing default points to the end of the waveform to meet
#             length constraints.
#         :param strict_hrdw_seq: (bool) [argument of pb_sample()]
#             if True, strict waveform length constraints will be checked
#             (necessary for Hardware Sequence to operate in Sequence run mode)

#         :return int: 0 - success code.
#                      PGenError, ValueError, and other exceptions are produced
#                      in the cases of errors.
#         """

#         # Get waveform length constraints
#         wfm_len_constr = self.get_wfm_len_constr(strict_hrdw_seq=strict_hrdw_seq)

#         # Sample pulse block
#         samp_dict, n_samp_pts, add_samp_pts = pb_sample(
#             pb_obj=pb_obj,
#             samp_rate=self.get_samp_rate(),
#             len_min=wfm_len_constr['min'],
#             len_max=wfm_len_constr['max'],
#             len_step=wfm_len_constr['step'],
#             len_adj=len_adj,
#             debug=False
#         )

#         # Write waveform to the AWG memory
#         self._write_wfm(
#             samp_dict=samp_dict,
#             wfm_name=pb_obj.name
#         )

#         return 0

#     def write_wfm_zip(self, pb_obj, len_adj=True):
#         """ Write waveform as a sub-sequence with wait periods
#         collapsed into repetitions of the same short wait waveform.

#         :param pb_obj: PulseBlock object to write
#         :param len_adj: (bool) if True, the waveform will be padded with
#             leading default-value points to meet waveform length constraints.
#             If False and length does not meet constraints (min, max, step),
#             ValueError will be produced. Note that if length exceeds max,
#             ValueError is always produced.

#         :return: 0 - success
#             PGenError, ValueError, or another exception is produced in the case
#             of error.
#         """

#         len_constr_dict = self.get_wfm_len_constr(strict_hrdw_seq=True)
#         len_min = len_constr_dict['min']
#         len_max = len_constr_dict['max']
#         len_step = len_constr_dict['step']

#         samp_rate = self.get_samp_rate()

#         # Zip (collapse) pulse block ------------------------------------------

#         dur_quant = len_min / samp_rate
#         zip_dict = pb_zip(
#             pb_obj=pb_obj,
#             dur_quant=dur_quant
#         )
#         self.log.debug(msg_str='write_wfm_zip(): completed pb_zip() call')

#         # Sample and write all individual waveform snippets -------------------

#         snip_list = zip_dict['snip_list']
#         for snip_idx in range(len(snip_list)):
#             # Only the last snippet can have length different from len_min
#             # - external argument len_adj should be passed when sampling it
#             # - no sanity check for len % len_min == 0 is required for it
#             is_last = (
#                 snip_idx == len(snip_list) - 1
#             )

#             samp_dict, n_samp_pts, _ = pb_sample(
#                 pb_obj=snip_list[snip_idx],
#                 samp_rate=samp_rate,
#                 len_min=len_min,
#                 len_max=len_max,
#                 len_step=len_step,
#                 len_adj=len_adj if is_last else False
#             )
#             # Sanity check:
#             #   all snippets (except the last one) are expected to be of len_min
#             #   (corresponding to dur_quant).
#             #   Only the last element can be length-adjusted
#             #   (where adjustment is equivalent to total wfm len adjustment)
#             if not is_last and n_samp_pts % len_min != 0:
#                 msg_str = 'write_wfm_zip(): sampling of pb_snip={} ' \
#                           'resulted in unexpected sample array length {}. \n' \
#                           'The expectation was integer multiple of the hardware min {}. \n' \
#                           'Having precise length is necessary for operation ' \
#                           'of hardware sequencer.' \
#                           ''.format(
#                               snip_list[snip_idx].name,
#                               n_samp_pts,
#                               len_min
#                           )
#                 self.log.error(msg_str=msg_str)
#                 raise PGenError(msg_str)

#             self._write_wfm(
#                 samp_dict=samp_dict,
#                 wfm_name=snip_list[snip_idx].name
#             )

#         self.log.debug(
#             msg_str='write_wfm_zip(): completed sampling and loading all '
#                     'individual waveform snippets'
#         )

#         # Construct sub-sequence ----------------------------------------------

#         if pb_obj.name in self.subseq_get_names():
#             self.subseq_del(name=pb_obj.name)

#         seq_len = len(zip_dict['seq_list'])
#         self.subseq_new(
#             name=pb_obj.name,
#             length=seq_len
#         )

#         self.log.debug(
#             msg_str='write_wfm_zip(): created new sub-sequence. \n'
#                     'Start filling-in sequence table'
#         )

#         # Fill-in sequence steps
#         for elem_idx in range(seq_len):
#             elem_wfm_name, elem_rep = zip_dict['seq_list'][elem_idx]

#             # Set waveforms
#             for a_ch_name in self._get_all_anlg_chs():
#                 self.subseq_set_wfm(
#                     subseq_name=pb_obj.name,
#                     elem_num=elem_idx + 1,
#                     a_ch_name=a_ch_name,
#                     wfm_name=elem_wfm_name + a_ch_name[1:]
#                 )

#             # Set repetition
#             self.subseq_set_rep(
#                 subseq_name=pb_obj.name,
#                 elem_num=elem_idx + 1,
#                 rep_num=elem_rep
#             )

#         return 0

#     def load_wfm(self, load_dict):
#         """Load wfm_name from 'User defined' list to specified channel of AWG.

#         :param load_dict: a dictionary with keys being one of the channel number
#                           and values being the name of the already written wfm_name
#                           to load into the channel.
#                           Examples:   {1: 'rabi_ch1', 2: 'rabi_ch2'}

#         :return dict: Dictionary containing the actually loaded waveforms per channel.
#         """

#         # Load waveforms into channels
#         for ch_num, wfm_name in load_dict.items():

#             # load into channel
#             self.write('SOUR{0:d}:WAV "{1}"'.format(ch_num, wfm_name))

#         return self.get_loaded_assets()

#     def get_wfm_names(self):
#         """ Retrieve the names on 'User defined' list.

#         :return list: 'User defined' list.
#         """

#         wfm_list_len = int(self.query('WLIS:SIZE?'))
#         wfm_list = list()
#         for index in range(wfm_list_len):
#             wfm_list.append(self.query('WLIS:NAME? {0:d}'.format(index)))
#         return sorted(wfm_list)

#     def del_wfm(self, wfm_name):
#         """ Delete waveform_name from 'User Defined' list.

#         :param str wfm_name: The name/list of names of the waveform to be deleted
#                            Optionally 'all' can be passed to delete all waveforms.

#         :return int: 0 - success code.
#                      PGenError is produced in the case of error.
#         """

#         # Delete all
#         if wfm_name == 'all':
#             self.write('WLIS:WAV:DEL ALL')
#             return 0

#         # Delete specified
#         else:
#             if isinstance(wfm_name, str):
#                 wfm_name = [wfm_name]
#             wfm_list = wfm_name

#             for wfm in wfm_list:
#                 self.write('WLIS:WAV:DEL "{0}"'.format(wfm))

#             return 0

#     # Waveform technical methods

#     def _calc_byte_ar(self, amp_pp, offset, anlg_ar, mrk1_ar, mrk2_ar):

#         v_min = offset - amp_pp/2
#         v_max = offset + amp_pp/2
#         v_step = amp_pp / (2**8 - 1)

#         # Sanity check: all values of anlg_ar are within DAC range
#         ar_min = np.amin(anlg_ar)
#         ar_max = np.amax(anlg_ar)

#         if not v_min <= ar_min:
#             msg_str = '_calc_byte_ar(): min requested value {} is below DAC min {}' \
#                       ''.format(ar_min, v_min)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         if not ar_max <= v_max:
#             msg_str = '_calc_byte_ar(): max requested value {} is above DAC max {}' \
#                       ''.format(ar_max, v_max)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         # Calculate DAC bits (closest integer number of v_step)
#         a_int_ar = (anlg_ar - v_min) // v_step

#         # Bit offset (see AWG docs)
#         a_offs = 2 ** 2
#         m1_offs = 2 ** 13
#         m2_offs = 2 ** 14

#         byte_ar = np.asarray(
#             a=mrk2_ar*m2_offs + mrk1_ar*m1_offs + a_int_ar*a_offs,
#             dtype=np.uint16
#         )

#         return byte_ar

#     def _write_wfm(self, samp_dict, wfm_name):

#         # Get analog levels (will be used to calculate DAC bits)
#         anlg_level_dict = self.get_analog_level()

#         # Write waveforms. One for each analog channel.
#         for a_ch_name in self._get_all_anlg_chs():

#             # Calculate bytes array -------------------------------------------

#             start_t = time.time()

#             # Get the analog channel number
#             a_ch_num = int(a_ch_name.rsplit('ch', 1)[1])
#             # Get the corresponding digital channel names
#             mrk1_name = 'd_ch{0:d}'.format(a_ch_num * 2 - 1)
#             mrk2_name = 'd_ch{0:d}'.format(a_ch_num * 2)

#             byte_ar = self._calc_byte_ar(
#                 amp_pp=anlg_level_dict[a_ch_name]['amp_pp'],
#                 offset=anlg_level_dict[a_ch_name]['offset'],
#                 anlg_ar=samp_dict[a_ch_name],
#                 mrk1_ar=samp_dict[mrk1_name],
#                 mrk2_ar=samp_dict[mrk2_name]
#             )

#             self.log.debug(
#                 'Calculated byte data: {:.3f} s'.format(time.time() - start_t)
#             )

#             # Write bytes array to local HDD file -----------------------------

#             start = time.time()

#             # Create waveform name string
#             file_name = '{0}_ch{1:d}'.format(wfm_name, a_ch_num)
#             # Write WFM to a file on local HDD
#             self._write_wfm_file(
#                 filename=file_name,
#                 byte_ar=byte_ar
#             )

#             self.log.debug(
#                 'Wrote WFM file to local HDD: {:.3f} s'.format(time.time() - start)
#             )

#             # Send file to AWG over FTP ---------------------------------------

#             start = time.time()
#             self._send_file(filename=file_name)
#             self.log.debug(
#                 'Sent WFM file to AWG HDD: {:.3f} s'.format(time.time() - start)
#             )

#             # Load waveform to the AWG fast memory ----------------------------
#             # (waveform will appear on "User Defined" list)

#             start = time.time()

#             # Set AWG current working dir to 'C:\\inetpub\\ftproot\\remote_wfm_dir'
#             self.write(
#                 'MMEM:CDIR "{0}"'
#                 ''.format(
#                     os.path.join('C:\\inetpub\\ftproot', self._remote_wfm_dir)
#                 )
#             )
#             self.write(
#                 'MMEM:IMP "{0}","{1}",PAT'
#                 ''.format(
#                     file_name,
#                     file_name + '.pat'
#                 )
#             )  # This operation can take about 10 s to complete for 64 MSa waveform

#             self.log.debug(
#                 'Loaded WFM file into "User Defined" list: {:.3f} s'
#                 ''.format(time.time() - start)
#             )

#         return 0

#     def _write_wfm_file(self, filename, byte_ar):
#         """ Write byte_ar to binary file filename

#         :param byte_ar: numpy array of np.uint16 data type
#         :param filename: string, represents the name of the sampled waveform
#         :return int: status code: 0 - success
#                      PGenError exception is produced in the case of error
#         """

#         total_samp_num = len(byte_ar)

#         # Create the WFM file.
#         if not filename.endswith('.pat'):
#             filename += '.pat'
#         wfm_path = os.path.join(self._local_wfm_dir, filename)

#         #   - write header
#         with open(wfm_path, 'wb') as wfm_file:
#             num_bytes = str(total_samp_num * 2)
#             num_digits = str(len(num_bytes))
#             header = 'MAGIC 2000\r\n#{0}{1}'.format(num_digits, num_bytes)
#             wfm_file.write(header.encode())

#         # Write byte_ar to file
#         with open(wfm_path, 'ab') as wfm_file:
#             wfm_file.write(byte_ar)

#         # Append the footer: the sample rate, which was used for that file
#         footer = 'CLOCK {0:16.10E}\r\n'.format(self.get_samp_rate())
#         with open(wfm_path, 'ab') as wfm_file:
#             wfm_file.write(footer.encode())

#         return 0

#     def _send_file(self, filename):
#         """ Send binary file filename over FTP

#         :param filename: file name
#         :return: status code: 0 - Ok
#                  Exception is produced in the case of error
#         """

#         if not filename.endswith('.pat'):
#             filename += '.pat'
#         filepath = os.path.join(self._local_wfm_dir, filename)

#         # Sanity check: file is present
#         if not os.path.isfile(filepath):
#             msg_str = '_send_file: No file "{0}" found in "{1}". Unable to send!'\
#                       ''.format(filename, self._local_wfm_dir)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         # Delete old file on AWG by the same filename
#         if filename in self._remote_dir_list():
#             self._del_remote_file(filename)

#         # Transfer file
#         with FTP(self._ftp_ip_str) as ftp:
#             ftp.login(user=self._ftp_username, passwd=self._ftp_pswrd)
#             ftp.cwd(self._remote_wfm_dir)
#             with open(filepath, 'rb') as file:  # In this case "file" refers to the file on user's PC
#                 ftp.storbinary('STOR ' + filename, file)

#         return 0

#     def _del_remote_file(self, filename):
#         """ Delete file from AWG HDD

#         File must be in 'C:\\inetpub\\ftproot\\remote_wfm_dir' directory

#         :param str filename: The full filename to delete from FTP cwd
#         :return int: status code: 0 -Ok
#         """

#         with FTP(self._ftp_ip_str) as ftp:
#             ftp.login(user=self._ftp_username, passwd=self._ftp_pswrd)
#             ftp.cwd(self._remote_wfm_dir)
#             ftp.delete(filename)

#         return 0

#     def _remote_dir_list(self):
#         """ Get list of all files in 'C:\\inetpub\\ftproot\\remote_wfm_dir' directory.

#         :return list: list of file name strings
#         """

#         filename_list = list()
#         with FTP(self._ftp_ip_str) as ftp:
#             ftp.login(user=self._ftp_username, passwd=self._ftp_pswrd)
#             ftp.cwd(self._remote_wfm_dir)
#             # get only the files from the dir and skip possible directories
#             log = list()
#             ftp.retrlines('LIST', callback=log.append)
#             for line in log:
#                 if '<DIR>' not in line:
#                     # that is how a potential line is looking like:
#                     #   '05-10-16  05:22PM                  292 SSR aom adjusted.seq'
#                     # The first part consists of the date information. Remove this information and
#                     # separate the first number, which indicates the size of the file. This is
#                     # necessary if the filename contains whitespaces.
#                     size_filename = line[18:].lstrip()
#                     # split after the first appearing whitespace and take the rest as filename.
#                     # Remove for safety all trailing and leading whitespaces:
#                     filename = size_filename.split(' ', 1)[1].strip()
#                     filename_list.append(filename)

#         return filename_list

#     def get_wfm_len_constr(self, strict_hrdw_seq=False):
#         """Get waveform length constraints

#         :param strict_hrdw_seq: whether strict hardware sequencing
#         is required (imposes strong length constraints, required for
#         using Sequnce mode to collapse large waveforms)

#         :return: (dict) constraints dictionary, depending on
#             strrict_hrdw_seq parapeter and interleave state:
#             {
#                 'min': minimal length,
#                 'max': maximal length,
#                 'step': granularity
#             }
#         """

#         # FIXME: this is not general.

#         # Constraints depend on whether interleave is enabled or not
#         interleave = self.get_interleave()

#         # Hardware sequencing is strictly required
#         # (for using Sequence run mode to collapse large waveforms)
#         if strict_hrdw_seq:
#             if interleave:
#                 return {
#                     'min': 1920,
#                     'max': int(19.5e6),
#                     'step': 8
#                 }
#             else:
#                 return {
#                     'min': 960,
#                     'max': int(19.5e6),
#                     'step': 4
#                 }

#         # No requirement to use Hardware Sequencer only
#         # (just to output a plain waveform without memory-saving sequencing)
#         # Constraints are much softer in this case
#         else:
#             if interleave:
#                 return {
#                     'min': 1,
#                     'max': int(19.5e6),
#                     'step': 1
#                 }
#             else:
#                 return {
#                     'min': 1,
#                     'max': int(19.5e6),
#                     'step': 1
#                 }

#     # -------------------------------------------------------------------------
#     # Sequence and sub-sequence
#     # -------------------------------------------------------------------------

#     # Basic

#     def seq_get_len(self):
#         return int(
#             self.query('SEQUENCE:LENGTH?')
#         )

#     def seq_set_len(self, seq_len):
#         self.write(
#             'SEQUENCE:LENGTH {}'.format(seq_len)
#         )

#         return self.seq_get_len()

#     def seq_get_wfm(self, elem_num, a_ch_name):
#         a_ch_num = int(
#             a_ch_name.rsplit('_ch', 1)[1]
#         )

#         return self.query(
#             'SEQUENCE:ELEMENT{}:WAVEFORM{}?'.format(
#                 elem_num,
#                 a_ch_num
#             )
#         )

#     def seq_set_wfm(self, elem_num, a_ch_name, wfm_name):
#         a_ch_num = int(
#             a_ch_name.rsplit('_ch', 1)[1]
#         )

#         self.write(
#             'SEQUENCE:ELEMENT{}:WAVEFORM{} "{}"'.format(
#                 elem_num,
#                 a_ch_num,
#                 wfm_name
#             )
#         )

#         return self.seq_get_wfm(
#             elem_num=elem_num,
#             a_ch_name=a_ch_name
#         )

#     def seq_set_subseq(self, elem_num, subseq_name):
#         self.write(
#             'SEQUENCE:ELEMENT{}:SUBSEQUENCE "{}"'.format(
#                 elem_num,
#                 subseq_name
#             )
#         )

#         return self.query(
#             'SEQUENCE:ELEMENT{}:SUBSEQUENCE?'.format(elem_num)
#         )

#     def seq_get_twait(self, elem_num):
#         return bool(int(
#             self.query(
#                 'SEQUENCE:ELEMENT{}:TWAIT?'.format(elem_num)
#             )
#         ))

#     def seq_set_twait(self, elem_num, trig_enable):
#         self.write(
#             'SEQUENCE:ELEMENT{}:TWAIT {}'.format(
#                 elem_num,
#                 int(trig_enable)
#             )
#         )

#         return self.seq_get_twait(elem_num=elem_num)

#     def seq_get_rep(self, elem_num):
#         """ Get repetition number of elem_num

#         :param elem_num: (int) sequence element number
#                          [1 corresponds to the first element]
#         :return: (int) 0 - infinite
#                        positive integer - number of repetitions
#         """

#         is_inf = bool(int(
#             self.query(
#                 'SEQUENCE:ELEMENT{}:LOOP:INFINITE?'.format(elem_num)
#             )
#         ))

#         if is_inf:
#             return 0
#         else:
#             return int(
#                 self.query(
#                     'SEQUENCE:ELEMENT{}:LOOP:COUNT?'.format(elem_num)
#                 )
#             )

#     def seq_set_rep(self, elem_num, rep_num):
#         """ Set repetition number of elem_num

#         :param elem_num: (int) sequence element number
#                          [1 corresponds to the first element]
#         :param rep_num: (int) 0 - infinite
#                         positive integer - number of repetitions
#         :return: (int) actual number of repetitions
#         """

#         if rep_num == 0:
#             self.write(
#                 'SEQUENCE:ELEMENT{}:LOOP:INFINITE 1'.format(elem_num)
#             )
#         elif rep_num > 0:
#             self.write(
#                 'SEQUENCE:ELEMENT{}:LOOP:INFINITE 0'.format(elem_num)
#             )
#             self.write(
#                 'SEQUENCE:ELEMENT{}:LOOP:COUNT {}'.format(
#                     elem_num,
#                     rep_num
#                 )
#             )
#         else:
#             msg_str = 'seq_set_rep(): invalid argument rep_num = {}. \n' \
#                       'Valid values: 0 - for inf, positive integer - for finite' \
#                       ''.format(rep_num)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return self.seq_get_rep(elem_num=elem_num)

#     def seq_get_goto(self, elem_num):
#         """Get go-to target for elem_num sequence element

#         :param elem_num: (int) number of seq element [first seq elem is 1]
#         :return: (int) actual target:
#                 0 - 'go to the next'
#                 positive integer - non-trivial target
#         """

#         is_enabled = bool(int(
#             self.query(
#                 'SEQUENCE:ELEMENT{}:GOTO:STATE?'.format(elem_num)
#             )
#         ))

#         if not is_enabled:
#             return 0
#         else:
#             return int(
#                 self.query(
#                     'SEQUENCE:ELEMENT{}:GOTO:INDEX?'.format(elem_num)
#                 )
#             )

#     def seq_set_goto(self, elem_num, target):
#         """Set go-to target for elem_num sequence element

#         :param elem_num: (int) number of seq element [first seq elem is 1]
#         :param target: (int) 0 - 'go to the next'
#                              positive integer - non-trivial target
#         :return: (int) actual target
#         """

#         if target == 0:
#             # Set trivial go-to (go to the next)
#             self.write(
#                 'SEQUENCE:ELEMENT{}:GOTO:STATE 0'.format(elem_num)
#             )

#         elif target > 0:
#             # Enable non-trivial go-to
#             self.write(
#                 'SEQUENCE:ELEMENT{}:GOTO:STATE 1'.format(elem_num)
#             )
#             # Set go-to target
#             self.write(
#                 'SEQUENCE:ELEMENT{}:GOTO:INDEX {}'.format(
#                     elem_num,
#                     target
#                 )
#             )

#         else:
#             msg_str = 'seq_set_goto(): invalid argument target {}. \n' \
#                       'Valid values are: \n' \
#                       '     0 - "go to the next" \n' \
#                       '     positive integer - target' \
#                       ''.format(target)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return self.seq_get_goto(elem_num=elem_num)

#     # Event / dynamic jump

#     def get_jmp_mode(self):
#         mode_str = self.query('AWGControl:EVENt:JMODe?')
#         if mode_str == 'EJUM':
#             return 'EJUMP'
#         elif mode_str == 'DJUM':
#             return 'DJUMP'
#         else:
#             msg_str = 'get_jmp_mode(): unknown jump mode string {} was returned' \
#                       ''.format(mode_str)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#     def set_jmp_mode(self, mode_str):
#         """ Set sequence jump mode: Event Jump or Dynamic Jump.

#         :param mode_str: 'EJUMP' - Event Jump
#                          'DJUMP' - Dynamic Jump
#         :return: actual jump mode string
#         """

#         self.write(
#             'AWGCONTROL:EVENT:JMODE {}'.format(mode_str)
#         )
#         return self.get_jmp_mode()

#     def get_jmp_timing(self):
#         return self.query('EVENT:JTIMING?')

#     def set_jmp_timing(self, sync=True):
#         if sync:
#             self.write('EVENT:JTIMING SYNCHRONOUS')
#         else:
#             self.write('EVENT:JTIMING ASYNCHRONOUS')

#         return self.get_jmp_timing()

#     def get_ejmp_trgt(self, elem_num):
#         """ Get Event Jump target (use get/set_djmp_trgt for Dynamic Jump mode)

#         :param elem_num: (int) sequence element number
#         (1 corresponds to the first sequence element)
#         :return: (int) target sequence element number
#                        -1 - ignore event signal for this element
#                         0 - jump to the next
#          positive integer - jump to the specified element
#         """

#         jmp_type_str = self.query(
#             'SEQUENCE:ELEMENT{}:JTARGET:TYPE?'.format(elem_num)
#         )

#         # Ignore event signal for this element
#         if jmp_type_str == 'OFF':
#             return -1

#         # Jump to the next
#         elif jmp_type_str == 'NEXT':
#             return 0

#         # Jump to the specified index
#         elif jmp_type_str == 'IND':
#             target_str = self.query(
#                 'SEQUENCE:ELEMENT{}:JTARGET:INDEX?'.format(elem_num)
#             )
#             return int(target_str)

#         # Unknown type
#         else:
#             msg_str = 'get_ejmp_trgt(): query for elem_num={} returned unknown jump type string {}' \
#                       ''.format(elem_num, jmp_type_str)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#     def set_ejmp_trgt(self, elem_num, target):
#         """ Set Event Jump target (use get/set_djmp_trgt for Dynamic Jump mode)

#         :param elem_num: (int) sequence element number
#         (1 corresponds to the first sequence element)

#         :param target: (int) target sequence element number
#                        -1 - ignore event signal for this element
#                         0 - jump to the next
#          positive integer - jump to the specified element

#         :return: (int) actual jump target.
#         """

#         # No Event jump action for this sequence element
#         if target < 0:
#             self.write(
#                 'SEQUENCE:ELEMENT{}:JTARGET:TYPE OFF'.format(elem_num)
#             )

#         # Jump to the next
#         elif target == 0:
#             self.write(
#                 'SEQUENCE:ELEMENT{}:JTARGET:TYPE NEXT'.format(elem_num)
#             )

#         # Jump to non-trivial index
#         else:
#             self.write(
#                 'SEQUENCE:ELEMENT{}:JTARGET:TYPE INDEX'.format(elem_num)
#             )
#             self.write(
#                 'SEQUENCE:ELEMENT{}:JTARGET:INDEX {}'.format(elem_num, target)
#             )

#         return self.get_ejmp_trgt(elem_num=elem_num)

#     def get_djmp_trgt(self, bit_patrn_int):
#         """ Get Dynamic Jump target for the specified bit pattern

#         :param bit_patrn_int: integer representation of bit patter:
#                               '0000 0101' is specified as 5
#         :return: (int) target element for this bit pattern
#         """

#         return int(self.query(
#             'AWGCONTROL:EVENT:DJUMP:DEFINE? {}'.format(
#                 bit_patrn_int
#             )
#         ))

#     def set_djmp_trgt(self, bit_patrn_int, trgt_elem):
#         """ Set Dynamic Jump target for the specified bit pattern

#         :param bit_patrn_int: integer representation of bit patter:
#                               '0000 0101' is specified as 5
#         :param trgt_elem: sequence element number to jump onto
#                           (1 corresponds to the first sequence element).
#                           Passing 0 disables dynamic jump for this bit
#                           pattern

#         :return: (int) actual target element for this bit pattern
#         """

#         self.write(
#             'AWGCONTROL:EVENT:DJUMP:DEFINE {},{}'.format(
#                 bit_patrn_int,
#                 trgt_elem
#             )
#         )
#         return self.get_djmp_trgt(bit_patrn_int=bit_patrn_int)

#     # Sub-sequence

#     def subseq_new(self, name, length):
#         self.write(
#             'SLIST:SUBSEQUENCE:NEW "{}",{}'.format(name, length)
#         )

#         return 0

#     def subseq_del(self, name):
#         """ Delete sub-sequence from sub-sequence list

#         :param name: (str) sub-sequence name or 'all'
#         :return: (int) 0 - success
#                  PGenError is produced in the case of error
#         """

#         if name == 'all':
#             self.write(
#                 'SLIST:SUBSEQUENCE:DELETE ALL'
#             )
#         else:
#             self.write(
#                 'SLIST:SUBSEQUENCE:DELETE "{}"'.format(name)
#             )

#         return 0

#     def subseq_get_len(self, name):
#         len_str = self.query(
#             'SLIST:SUBSEQUENCE:LENGTH? "{}"'.format(name)
#         )
#         return int(len_str)

#     def subseq_set_len(self, name, length):
#         self.write(
#             'SLISt:SUBSequence:LENGth "{}",{}'.format(
#                 name,
#                 length
#             )
#         )

#         return self.subseq_get_len(name=name)

#     def subseq_get_wfm(self, subseq_name, elem_num, a_ch_name):
#         a_ch_num = int(
#             a_ch_name.rsplit('_ch', 1)[1]
#         )

#         return self.query(
#             'SLIST:SUBSEQUENCE:ELEMENT{}:WAVEFORM{}? "{}"'.format(
#                 elem_num,
#                 a_ch_num,
#                 subseq_name
#             )
#         )

#     def subseq_set_wfm(self, subseq_name, elem_num, a_ch_name, wfm_name):
#         a_ch_num = int(
#             a_ch_name.rsplit('_ch', 1)[1]
#         )

#         self.write(
#             'SLIST:SUBSEQUENCE:ELEMENT{}:WAVEFORM{} "{}","{}"'.format(
#                 elem_num,
#                 a_ch_num,
#                 subseq_name,
#                 wfm_name
#             )
#         )

#         return self.subseq_get_wfm(
#             subseq_name=subseq_name,
#             elem_num=elem_num,
#             a_ch_name=a_ch_name
#         )

#     def subseq_get_rep(self, subseq_name, elem_num):
#         rep_num = int(self.query(
#             'SLIST:SUBSEQUENCE:ELEMENT{}:LOOP:COUNT? "{}"'.format(
#                 elem_num,
#                 subseq_name
#             )
#         ))

#         return rep_num

#     def subseq_set_rep(self, subseq_name, elem_num, rep_num):
#         self.write(
#             'SLIST:SUBSEQUENCE:ELEMENT{}:LOOP:COUNT "{}",{}'.format(
#                 elem_num,
#                 subseq_name,
#                 rep_num
#             )
#         )

#         return self.subseq_get_rep(
#             subseq_name=subseq_name,
#             elem_num=elem_num
#         )

#     def subseq_get_names(self):

#         ss_name_list = []

#         # Number of seb-sequences on the list
#         ss_list_len = int(
#             self.query('SLIST:SIZE?')
#         )

#         for idx in range(1, ss_list_len + 1):
#             ss_name_list.append(
#                 self.query('SLIST:NAME? {}'.format(idx))
#             )

#         return ss_name_list

#     # -------------------------------------------------------------------------
#     # Waveform and sequence technical
#     # -------------------------------------------------------------------------

#     def get_loaded_assets(self):
#         """
#         Retrieve the currently loaded asset names for each active channel of the device.
#         The returned dictionary will have the channel numbers as keys.
#         In case of loaded waveforms the dictionary values will be the waveform names.
#         In case of a loaded sequence the values will be the sequence name appended by a suffix
#         representing the track loaded to the respective channel (i.e. '<sequence_name>_1').

#         @return (dict, str): Dictionary with keys being the channel number and values being the
#                              respective asset loaded into the channel,
#                              string describing the asset type ('waveform' or 'sequence')
#         """

#         # Get all active analog channels
#         a_ch_name_list = self._get_all_anlg_chs()
#         a_ch_num_list = sorted(
#             int(ch_name.split('_ch')[1]) for ch_name in a_ch_name_list
#         )

#         # Get assets per channel
#         loaded_assets = dict()

#         run_mode = self.query('AWGC:RMOD?')
#         if run_mode in ['CONT', 'TRIG', 'GAT']:
#             current_type = 'waveform'
#             for ch_num in a_ch_num_list:
#                 loaded_assets[ch_num] = self.query('SOUR{0}:WAV?'.format(ch_num))
#         elif run_mode == 'SEQ':
#             current_type = 'sequence'
#             for ch_num in a_ch_num_list:
#                 if len(self._loaded_seqs) > 0:
#                     loaded_assets[ch_num] = self._loaded_seqs[0]
#         else:
#             msg_str = 'get_loaded_assets(): received unknown run mode: {}'\
#                       ''.format(run_mode)
#             self.log.error(msg_str=msg_str)
#             raise PGenError(msg_str)

#         return loaded_assets, current_type

#     def clear_all(self):
#         """ Clears all loaded waveforms from the pulse generators RAM/workspace.
#         (from "User Defined" list)

#         @return int: error code (0:OK, -1:error)
#         """

#         self.write('WLIS:WAV:DEL ALL')
#         if '09' in self.option_list:
#             self.write('SLIS:SUBS:DEL ALL')

#         self.write('SEQUENCE:LENGTH 0')
#         self._written_seqs = []
#         self._loaded_seqs = []

#         return 0


# class Service(ServiceBase):
#     # TODO: implement
#     pass


# class Client(ClientBase):
#     # TODO: implement
#     pass



import socket

IMPORT_STATUS = True

try:
    import TimeTagger as TT
except ModuleNotFoundError:
    IMPORT_STATUS = False

from pylabnet.hardware.counter.swabian_instruments.cnt_monitor import Wrap
from pylabnet.network.core.generic_server import GenericServer
from pylabnet.network.client_server.si_tt_cnt_monitor import Service, Client


def launch(**kwargs):
    """ Connects to SI TT and instantiates server

    :param kwargs: (dict) containing relevant kwargs
        :logger: instance of LogClient for logging purposes
        :port: (int) port number for the Cnt Monitor server
    """

    if not IMPORT_STATUS:
        msg_str = 'Please make sure Swabian Instruments drivers are installed on this machine.'
        raise ModuleNotFoundError(msg_str)

    DEFAULT_CH_LIST = [1]

    TT.setTimeTaggerChannelNumberScheme(TT.TT_CHANNEL_NUMBER_SCHEME_ONE)

    # Connect to the device, otherwise instantiate virtual connection
    try:
        tagger = TT.createTimeTagger()
    except RuntimeError:
        kwargs['logger'].warn('Failed to connect to Swabian Instruments Time Tagger.'
                              ' Instantiating virtual device instead')
        tagger = TT.createTimeTaggerVirtual()

    cnt_trace_wrap = Wrap(
        tagger=tagger,
        ch_list=DEFAULT_CH_LIST,
        logger=kwargs['logger']
    )

    # Instantiate Server
    cnt_trace_service = Service()
    cnt_trace_service.assign_module(module=cnt_trace_wrap)
    cnt_trace_service.assign_logger(logger=kwargs['logger'])
    cnt_trace_server = GenericServer(
        service=cnt_trace_service,
        host=socket.gethostbyname(socket.gethostname()),
        port=kwargs['port']
    )
    cnt_trace_server.start()



""" Set of tools to stop continuously running script from a different process
via RPyC link.

Steps:
- the script class should have public method pause(),
  which will be called by client through RPyC link
- once the script is instantiated, assign it as the module to
  PauseService instance:
      pause_service_instance.assign_module(script_instance)
- instantiate PauseClient in a different process
- call of PauseClient.pause() will call pause() method of the script
"""

from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class PauseService(ServiceBase):
    """Makes PauseFlag instance visible on pylabnet network.

    Once the script is instantiated, assign it as the module to
    PauseService instance:
        pause_service_instance.assign_module(script_instance)
    """

    def exposed_pause(self):

        if isinstance(self._module, list):
            for module in self._module:
                module.pause()
            return 0

        else:
            return self._module.pause()


class PauseClient(ClientBase):
    """Client to send stop request to the script.

    Call of client_instance.pause() will call pause() method of the script
    """

    def pause(self):
        return self._service.exposed_pause()



# -*- coding: utf-8 -*-

"""
This file contains the pylabnet Hardware module class for a generic NI DAQ mx card.
"""

import nidaqmx
from pylabnet.hardware.interface.gated_ctr import GatedCtrInterface

from pylabnet.utils.logging.logger import LogHandler
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase

import pickle


class Driver:
    """Driver for NI DAQmx card. Currently only implements setting AO voltage"""
    # TODO implement counter

    def __init__(self, device_name, logger=None):
        """Instantiate NI DAQ mx card

        :device_name: (str) Name of NI DAQ mx card, as displayed in the measurement and automation explorer
        """

        # Device name
        self.dev = device_name

        # Log
        self.log = LogHandler(logger=logger)

        # Try to get info of DAQ device to verify connection
        try:
            ni_daq_device = nidaqmx.system.device.Device(name=device_name)
            self.log.info(
                "Successfully connected to NI DAQ '{device_name}' (type: {product_type}) \n"
                "".format(
                    device_name=ni_daq_device.name,
                    product_type=ni_daq_device.product_type
                )
            )

        # If failed, provide info about connected DAQs
        except nidaqmx.DaqError as exc_obj:

            # Log exception message

            # - get names of all connected NI DAQs
            ni_daqs_names = nidaqmx.system._collections.device_collection.\
                DeviceCollection().device_names

            # - exception message
            self.log.exception(
                "NI DAQ card {} not found. \n"
                "There are {} NI DAQs available: \n "
                "    {}"
                "".format(
                    device_name,
                    len(ni_daqs_names),
                    ni_daqs_names
                )
            )

            # Raise exception
            raise exc_obj

    def set_ao_voltage(self, ao_channel, voltages):
        """Set analog output of NI DAQ mx card to a series of voltages

        :ao_channel: (str) Name of output channel (e.g. 'ao1', 'ao2')
        :voltages: (list of int) list of voltages which will be output
        """

        # TODO: Understand the timing between output voltages (sample-wise?)
        channel = self._gen_ch_path(ao_channel)

        with nidaqmx.Task() as task:
            task.ao_channels.add_ao_voltage_chan(channel)
            task.write(voltages, auto_start=True)

    # Technical methods

    def _gen_ch_path(self, channel):
        """ Auxiliary method to build channel path string.

        :param channel: (str) channel name ['ao1']
        :return: (str) full channel name ['Dev1/ao1']
        """

        return "{device_name}/{channel}".format(
            device_name=self.dev,
            channel=channel
        )


class GatedCounter(GatedCtrInterface):
    """ Hardware class for NI gated counter """

    def __init__(self):
        self.ci_channel = None
        self.task = None
        self.duration = 0.1
        self._status = 'Inactive'
        self.current_count = 0

    def activate_interface(self, counter_channel='Dev1/ctr0', physical_channel='/Dev1/20MHzTimebase'):
        """ Activates counter interface (creates a task, does not start it)

        :param counter_channel: (str) channel of counter to use, e.g. 'Dev1/ctr0'
        :param physical_channel: (str) channel of physical counter input, e.g. 'Dev1/PFI0'
        """

        # Create a task - note we have to be careful and close the task if something goes wrong
        self.task = nidaqmx.Task()

        try:
            self.ci_channel = self.task.ci_channels.add_ci_count_edges_chan(counter_channel)
            self.ci_channel.ci_count_edges_term = physical_channel
            self._status = 'Active, but not counting'

        except nidaqmx.DaqError:
            self.task.close()
            self._status = 'Inactive'
            print('Failed to activate counter')
            raise

    def init_ctr(self, bin_number=0.1, gate_type=None):
        """ Initializes gated counter parameters

        :param bin_number: (double) number of seconds to read for
        :param gate_type: (not used)
        """
        self.duration = bin_number
        pass

    def close_ctr(self):
        """ Stops the task and closes it.

        The interface must be reactivated using activate_interface command in order to resume counting """

        try:
            self.task.stop()
        except nidaqmx.DaqError:
            pass
        self._status = 'Inactive'
        self.task.close()

    def start_counting(self):
        """ Starts the counter """

        self._status = 'Counting'
        try:
            self.task.start()
            self.current_count = self.task.read()
        except nidaqmx.DaqError:
            pass

    def terminate_counting(self):
        pass

    def get_status(self):
        """ Returns status of the counter

        :return: (str) whether or not the counter is active, and counting
        """

        return self._status

    def get_count_ar(self, timeout=-1):
        pass
